"""Create tables from data-objects"""
# Author: Christian Brodbeck <christianbrodbeck@nyu.edu>
from itertools import zip_longest
from operator import itemgetter
import re
from typing import Callable, Sequence, Union
from warnings import warn

import numpy as np

from . import fmtxt
from ._celltable import Celltable
from ._exceptions import KeysMissing
from ._data_obj import (
    CategorialArg, IndexArg, VarArg,
    Categorial, Dataset, Factor, Interaction, NDVar, Scalar, UTS,
    Var, ascategorial, as_legal_dataset_key, asndvar, asvar, assub, asuv,
    cellname, combine, isuv)


def difference(y, x, c1, c0, match, sub=None, ds=None, by=None):
    """Subtract data in one cell from another

    Parameters
    ----------
    y : Var | NDVar
        Dependent variable.
    x : categorial
        Model for subtraction, providing categories to compute ``c1 - c0``.
    c1 : str | tuple
        Name of the cell in ``x`` that forms the minuend.
    c0 : str | tuple
        Name of the cell in ``x`` that is to be subtracted from ``c1``.
    match : categorial
        Units over which measurements were repeated. ``c1 - c0`` will be
        calculated separately for each level of ``match`` (e.g. ``"subject"``,
        or ``"subject % condition"``).
    sub : None | index
        Only include a subset of the data.
    ds : Dataset
        If a Dataset is specified other arguments can be str instead of
        data-objects and will be retrieved from ``ds``.

    Returns
    -------
    diff : Dataset
        Dataset with the difference between ``c1`` and ``c0`` on ``y``.

    Examples
    --------
    ERP difference wave: assuming a dataset ``ds`` with EEG data
    (``ds['eeg']``), a variable named ``'condition'`` with levels ``'expected'``
    and ``'unexpected'``, and multiple subjects, the following will generate the
    ``unexpected - expected`` difference waves::

        >>> diff = table.difference('eeg', 'condition', 'unexpected', 'expected',
        ... 'subject', ds=ds)

    If ``ds`` also contains a different factor crossed with ``condition``,
    called ``'word'`` with levels ``'verb'`` abd ``'adjective'``, then separate
    difference waves for verbs and adjectives can be computed with::

        >>> diff = table.difference('eeg', 'condition', 'unexpected', 'expected',
        ... 'subject % word', ds=ds)

    Given the latter, the difference of the difference waves could be computed
    with::

        >>> diffdiff = table.difference('eeg', 'word', 'verb', 'adjective',
        ... 'subject', ds=diff)

    """
    sub = assub(sub, ds)
    x = ascategorial(x, sub, ds)
    out = Dataset()
    if by is None:
        ct = Celltable(y, x, match, sub, ds=ds)
        if not ct.all_within:
            raise ValueError("Design is not fully balanced")
        groups = ct.groups[c1]
        if isinstance(groups, Interaction):
            for x in groups.base:
                out.add(x)
        else:
            out.add(groups)
        yname = y if isinstance(y, str) else ct.y.name
        out[yname] = ct.data[c1] - ct.data[c0]
        # Transfer other variables in ds that are compatible with the rm-structure
        if ds is not None:
            out.update(ct._align_ds(ds, True, out.keys(), isuv))
    else:
        warn("The by parameter is deprecated; use match instead", DeprecationWarning)
        by = ascategorial(by, sub, ds)
        ct = Celltable(y, x % by, match, sub, ds=ds)
        if not ct.all_within:
            raise ValueError("Design is not fully balanced")

        yname = y if isinstance(y, str) else ct.y.name
        if isinstance(c1, str):
            c1 = (c1,)
        if isinstance(c0, str):
            c0 = (c0,)
        dss = []
        ds_keep = None
        for cell in by.cells:
            if isinstance(cell, str):
                cell = (cell,)
            cell_ds = Dataset()
            cell_ds.add(ct.groups[c1 + cell])
            cell_ds[yname] = ct.data[c1 + cell] - ct.data[c0 + cell]
            if isinstance(by, Factor):
                cell_ds[by.name, :] = cell[0]
            else:
                for b, c in zip(by.base, cell):
                    cell_ds[b.name, :] = c
            # Transfer other variables in ds that are compatible with the rm-structure
            if ds_keep is None:
                if ds is None:
                    ds_keep = False
                else:
                    ds_keep = ct._align_ds(ds, True, cell_ds.keys(), isuv)
            if ds_keep:
                cell_ds.update(ds_keep)
            dss.append(cell_ds)
        out = combine(dss)

    return out


def frequencies(y, x=None, of=None, sub=None, ds=None):
    """Calculate frequency of occurrence of the categories in ``y``

    Parameters
    ----------
    y : univariate
        Values whose frequencies are of interest.
    x : categorial
        Optional model defining cells for which frequencies are displayed
        separately.
    of : categorial
        With ``x`` constant within ``of``, only count frequencies for each value
        in ``of`` once. (Compress y and x before calculating frequencies.)
    sub : index
        Only use a subset of the data.
    ds : Dataset
        If ds is specified, other parameters can be strings naming for
        variables in ``ds``.

    Returns
    -------
    freq : Dataset
        Dataset with frequencies.

    Examples
    --------
    A simple sample dataset::

        >>> ds = Dataset()
        >>> ds['a'] = Factor('aabbcc')
        >>> ds['x'] = Factor('xxxyyy')

    Display frequency of a single factor's cells::

        >>> print(table.frequencies('a', ds=ds))
        cell   n
        --------
        a      2
        b      2
        c      2

    Display frequency of interaction cells::

        >>> print(table.frequencies('a', 'x', ds=ds))
        x   a   b   c
        -------------
        x   2   1   0
        y   0   1   2

    """
    sub = assub(sub, ds)
    y = asuv(y, sub, ds, interaction=True)
    if x is not None:
        x = ascategorial(x, sub, ds)
    if of is not None:
        of = ascategorial(of, sub, ds)
        y = y.aggregate(of)
        if x is not None:
            x = x.aggregate(of)

    name = "Frequencies of %s" % (y.name,) if y.name else "Frequencies"
    if isinstance(y, Var):
        cells = np.unique(y.x)
    else:
        cells = y.cells

    # special case
    if x is None:
        out = Dataset(name=name)
        if isinstance(y, Interaction):
            for i, f in enumerate(y.base):
                out[f.name] = Factor((c[i] for c in x.cells),
                                     random=getattr(f, 'random', False))
        elif isinstance(y, Factor):
            out[y.name or 'y'] = Factor(cells, random=y.random)
        elif isinstance(y, Var):
            out[y.name or 'y'] = Var(cells)
        else:
            raise RuntimeError("y=%r" % (y,))
        n = np.fromiter((np.sum(y == cell) for cell in cells), int, len(cells))
        out['n'] = Var(n)
        return out

    # header
    if getattr(x, 'name', None):
        name += ' by %s' % x.name
    out = Dataset(name=name)

    if isinstance(x, Interaction):
        for i, f in enumerate(x.base):
            random = getattr(f, 'random', False)
            out[f.name] = Factor((c[i] for c in x.cells), random=random)
    else:
        out[x.name] = Factor(x.cells)

    y_idx = {cell: y == cell for cell in cells}
    x_idx = {cell: x == cell for cell in x.cells}
    for y_cell in cells:
        n = (np.sum(np.logical_and(y_idx[y_cell], x_idx[x_cell]))
             for x_cell in x.cells)
        name = as_legal_dataset_key(cellname(y_cell, '_'))
        out[name] = Var(np.fromiter(n, int, len(x.cells)))

    return out


def melt(name, cells, cell_var_name, ds, labels=None):
    """
    Restructure a Dataset such that a measured variable is in a single column

    Restructure a Dataset with a certain variable represented in several
    columns into a longer dataset in which the variable is represented in a
    single column along with an identifying variable.

    Additional variables are automatically included.

    Parameters
    ----------
    name : str
        Name of the variable in the new Dataset.
    cells : sequence of str | str
        Names of the columns representing the variable in the input Dataset.
        Names can either pe specified explicitly as a sequence of str, or
        implicitly as a str containing '%i' for an integer.
    cell_var_name : str
        Name of the variable to contain the cell identifier.
    ds : Dataset
        Input Dataset.
    labels : dict | sequence of str
        Labels for the keys in ``cells``. Can be specified either as a
        ``{key: label}`` dictionary or as a list of :class:`str` corresponding
        to ``cells``.

    Examples
    --------
    Simple example data::

        >>> ds = Dataset()
        >>> ds['y1'] = Var([1, 2, 3])
        >>> ds['y2'] = Var([4, 5, 6])
        >>> print(ds)
        y1   y2
        -------
        1    4
        2    5
        3    6
        >>> print(table.melt('y', ['y1', 'y2'], 'id', ds))
        y   id
        ------
        1   y1
        2   y1
        3   y1
        4   y2
        5   y2
        6   y2

    Additional variables are automatically included::

        >>> ds['rm'] = Factor('abc')
        >>> print(ds)
        y1   y2   rm
        ------------
        1    4    a
        2    5    b
        3    6    c
        >>> print(table.melt('y', ['y1', 'y2'], 'id', ds))
        rm   y   id
        -----------
        a    1   y1
        b    2   y1
        c    3   y1
        a    4   y2
        b    5   y2
        c    6   y2

    """
    # find source cells
    if isinstance(cells, str):
        cell_expression = cells
        cells = []
        cell_values = []
        if '%i' in cell_expression:
            pattern = cell_expression.replace('%i', r'(\d+)')
            for key in ds:
                m = re.match(pattern, key)
                if m:
                    cells.append(key)
                    cell_values.append(int(m.group(1)))
        else:
            raise ValueError(f"cells={cells!r}; If specified as string, it needs to contain '%i' as a place-holder for an integer that identifies columns")
    else:
        cell_values = cells

    if labels is None:
        cell_labels = cell_values
    elif isinstance(labels, dict):
        cell_labels = [labels[v] for v in cell_values]
    elif len(labels) != len(cells):
        raise ValueError(f"labels={labels!r}: needs as many entries as there are cells ({len(cells)})")
    else:
        cell_labels = labels

    # melt the Dataset
    keep = [k for k, v in ds.items() if isuv(v)]
    if keep:
        out = ds[tuple(keep)].tile(len(cells))
    else:
        out = Dataset(info=ds.info)
    out[name] = combine(ds[cell] for cell in cells)
    out[cell_var_name] = Factor(cell_labels, repeat=ds.n_cases)
    return out


def melt_ndvar(ndvar, dim=None, cells=None, ds=None, varname=None, labels=None):
    """
    Transform data to long format by converting an NDVar dimension into a variable

    Parameters
    ----------
    ndvar : NDVar | str
        The NDVar (or name of the NDVar in ``ds``).
    dim : str
        The name of the dimension that should be unwrapped (optional if
        ``ndvar`` has only one non-case dimension).
    cells : sequence
        The values on ``dim`` that should be included in the output Dataset
        (default is to include all values).
    ds : Dataset
        Dataset with additional variables that should be included in the long
        table.
    varname : str
        Name for the transformed ``ndvar`` (default is ``ndvar.name``).
    labels: dict | callable
        Mapping or function to create labels for ``dim`` levels.

    Returns
    -------
    long_table : Dataset
        Dataset in long format.

    See Also
    --------
    cast_to_ndvar

    Examples
    --------
    See :ref:`exa-compare-topographies`.
    """
    ndvar = asndvar(ndvar, ds=ds)
    if dim is None:
        if ndvar.ndim == ndvar.has_case + 1:
            dim = ndvar.dims[-1]
            dimname = dim.name
        else:
            raise ValueError(f"The ndvar has more than one possible dimensions, the dim parameter must be one of {ndvar.dimnames[ndvar.has_case:]}")
    else:
        dimname = dim
        dim = ndvar.get_dim(dimname)

    dim_vars = dim._melt_vars()
    if cells is None:
        cells = dim._as_uv()
    elif dim_vars:
        index = dim._array_index(cells)
        dim_vars = {k: v[index] for k, v in dim_vars.items()}

    if callable(labels):
        label = labels
    elif isinstance(labels, dict):
        missing = set(cells).difference(labels)
        if missing:
            raise KeysMissing(missing, 'labels', labels)
        label = itemgetter(labels)
    elif labels is None:
        label = lambda x: x
    else:
        raise TypeError(f"labels={labels!r}")

    if varname is None:
        if ndvar.name is None:
            raise TypeError("Need to provide a name")
        varname = ndvar.name

    if ds is None:
        base_ds = Dataset()
    else:
        uv_keys = tuple(k for k, v in ds.items() if isuv(v))
        base_ds = ds[uv_keys]

    if not ndvar.has_case:
        ndvar = ndvar[np.newaxis]
    dss = []
    for cell in cells:
        ds_ = base_ds.copy()
        ds_[varname] = ndvar.sub(**{dimname: cell})
        ds_[dimname, :] = label(cell)
        dss.append(ds_)
    out = combine(dss)
    for k, v in dim_vars.items():
        if k not in out:
            out[k] = v.repeat(len(ndvar))
    return out


def cast_to_ndvar(
        data: Union[VarArg, Sequence[VarArg]],
        dim_values: Union[VarArg, Factor],
        match: CategorialArg,
        sub: IndexArg = None,
        ds: Dataset = None,
        dim: str = None,
        unit: str = 's',
        name: str = None,
) -> Dataset:
    """Create an NDVar by converting a data column to a dimension
    
    Parameters
    ----------
    data
        Data to be cast.
    dim_values
        Location on the new dimension.
    match
        Indicating rows which belong the the same case in the NDvar.
    sub
        Use a subset of the data.
    ds
        Dataset with data for operation.
    dim
        Name for the new dimension. Use ``dim='uts'`` to create :class:`UTS` 
        time dimension from scalar ``dim_values``.
    unit
        Unit for :class:`UTS` dimension (ignored otherwise).
    name
        Name for the new :class:`NDVar` (the default is the name of ``data``).
        
    Returns
    -------
    short_ds
        Copy of ``ds``, aggregated over ``dim_values``, and with an 
        :class:`NDVar` containing the values form ``data`` and a new dimension
        reflecting ``dim_values``. If ``dim_values`` is a Factor, the new 
        dimension is :class:`Categorial`; if ``dim_values`` is a :class:`Var`, 
        it is :class:`Scalar`. The new dimension's name is ``dim``. The only
        exception to this is that when ``dim='uts'``, the new dimension is 
        :class:`UTS` named ``'time'``.

    See Also
    --------
    melt_ndvar
    """
    sub, n = assub(sub, ds, return_n=True)
    if isinstance(data, (str, Var)):
        data, n = asvar(data, sub, ds, n, return_n=True)
        data_vars = [data]
        names = [name]
    else:
        data_vars = []
        for data_i in data:
            data_var, n = asvar(data_i, sub, ds, n, return_n=True)
            data_vars.append(data_var)
        if name is None:
            names = [None for _ in range(len(data_vars))]
        elif isinstance(name, str):
            raise TypeError(f"name={name!r}: single name for multiple variables")
        else:
            names = name
    dim_values, n = asuv(dim_values, sub, ds, n, return_n=True)
    match, n = ascategorial(match, sub, ds, n, return_n=True)

    # determine NDVar dimension
    if isinstance(dim_values, Factor):
        unique_dim_vales = dim_values.cells
        dim = Categorial(dim or dim_values.name, unique_dim_vales)
    else:
        unique_dim_vales = np.unique(dim_values.x)
        if dim == 'uts':
            diff = np.diff(unique_dim_vales)
            unique_diff = np.unique(diff)
            if len(unique_diff) > 1:
                if np.diff(unique_diff).max() > 1e-15:
                    raise NotImplementedError(f"Can't create UTS dimension from data with irregular sampling (detected time-steps of {', '.join(map(str, unique_diff))}")
                tstep = round(unique_diff.mean(), 17)
            else:
                tstep = unique_diff[0]
            dim = UTS(unique_dim_vales[0], tstep, len(unique_dim_vales), unit)
        else:
            dim = Scalar(dim or dim_values.name, unique_dim_vales)

    # find NDVar data
    n_samples = len(dim)
    n_cases = len(match.cells)
    case_indexes = [match == case for case in match.cells]
    samples_indexes = [dim_values == v for v in unique_dim_vales]
    xs = [np.empty((n_cases, n_samples)) for _ in data_vars]
    index = None
    for i, case_index in enumerate(case_indexes):
        for j, sample_index in enumerate(samples_indexes):
            index = np.logical_and(case_index, sample_index, out=index)
            for x, data_var in zip(xs, data_vars):
                x[i, j] = data_var.x[index]

    # package output dataset
    if ds is None:
        out = Dataset()
    else:
        out = ds if sub is None else ds.sub(sub)
        out = out.aggregate(match, drop_bad=True, count=False)
    for x, data_var, name in zip_longest(xs, data_vars, names):
        out[name or data_var.name] = NDVar(x, ('case', dim))
    return out


def stats(
        y: Var,
        row: CategorialArg,
        col: CategorialArg = None,
        match: CategorialArg = None,
        sub: IndexArg = None,
        fmt: str = '%.4g',
        funcs: Sequence[Callable] = (np.mean,),
        ds: Dataset = None,
        title: fmtxt.FMTextArg = None,
        caption: fmtxt.FMTextArg = None,
):
    """Make a table with statistics

    Parameters
    ----------
    y
        Dependent variable.
    row
        Model specifying rows
    col
        Model specifying columns.
    match
        Identifier for repeated measures data; aggregate within subject before
        computing statistics.
    sub
        Only use part of the data.
    fmt
        How to format values.
    funcs
        A list of statistics functions to show (all functions must take an
        array argument and return a scalar).
    ds
        If a Dataset is provided, ``y``, ``row``, and ``col`` can be strings
        specifying members.
    title
        Table title.
    caption : str | FMText
        Table caption.


    Returns
    -------
    table : fmtxt.Table
        Table with statistics.

    Examples
    --------
    >>> ds = datasets.get_uts()
    >>> table.stats('Y', 'A', 'B', ds=ds)
                B
         -----------------
         b0        b1
    ----------------------
    a0   0.1668    -0.3646
    a1   -0.4897   0.8746

    >>> table.stats('Y', 'A', ds=ds, funcs=[np.mean, np.std])
    Condition   Mean     Std
    --------------------------
    a0          0.6691   1.37
    a1          0.8596   1.192

    """
    sub = assub(sub, ds)
    y = asvar(y, sub, ds)
    row = ascategorial(row, sub, ds)
    if match is not None:
        match = ascategorial(match, sub, ds)

    if col is None:
        ct = Celltable(y, row, match=match)

        # table header
        n_disp = len(funcs)
        table = fmtxt.Table('l' * (n_disp + 1), title=title, caption=caption)
        table.cell('Condition', 'bf')
        for func in funcs:
            table.cell(func.__name__.capitalize(), 'bf')
        table.midrule()

        # table entries
        for cell in ct.cells:
            data = ct.data[cell]
            table.cell(cell)
            for func in funcs:
                table.cell(fmt % func(data.x))
    else:
        col = ascategorial(col, sub, ds)
        ct = Celltable(y, row % col, match=match)

        N = len(col.cells)
        table = fmtxt.Table('l' * (N + 1), title=title, caption=caption)

        # table header
        table.cell()
        table.cell(col.name, width=N, just='c')
        table.midrule(span=(2, N + 1))
        table.cell()

        table.cells(*col.cells)
        table.midrule()

        # table body
        fmt_n = fmt.count('%')
        if fmt_n == 1:
            fmt_once = False
        elif len(funcs) == fmt_n:
            fmt_once = True
        else:
            raise ValueError("fmt does not match funcs")

        for Ycell in row.cells:
            table.cell(Ycell)
            for Xcell in col.cells:
                # construct address
                a = ()
                if isinstance(Ycell, tuple):
                    a += Ycell
                else:
                    a += (Ycell,)
                if isinstance(Xcell, tuple):
                    a += Xcell
                else:
                    a += (Xcell,)

                # cell
                data = ct.data[a]
                values = (f(data.x) for f in funcs)
                if fmt_once:
                    txt = fmt % values
                else:
                    txt = ', '.join((fmt % v for v in values))

                table.cell(txt)

    return table


def repmeas(y, x, match, sub=None, ds=None):
    """Create a repeated-measures table

    Parameters
    ----------
    y :
        Dependent variable (can be model with several dependents).
    x : categorial
        Model defining the cells that should be restructured into variables.
    match : categorial
        Model identifying the source of the measurement across repetitions,
        i.e. the model that should be retained.
    sub :
        boolean array specifying which values to include (generate e.g.
        with 'sub=T==[1,2]')
    ds : Dataset
        If a Dataset is specified other arguments can be str instead of
        data-objects and will be retrieved from ds.

    Returns
    -------
    rm_table : Dataset
        Repeated measures table. Entries for cells of ``x`` correspond to the
        data in ``y`` on these levels of ``x`` (if cell names are not valid
        Dataset keys they are modified).

    Examples
    --------
    Generate test data in long format::

        >>> ds = Dataset()
        >>> ds['y'] = Var([1,2,3,5,6,4])
        >>> ds['x'] = Factor('aaabbb')
        >>> ds['rm'] = Factor('123231', random=True)
        >>> print(ds)
        y   x   rm
        ----------
        1   a   1
        2   a   2
        3   a   3
        5   b   2
        6   b   3
        4   b   1

    Compute difference between two conditions::

        >>> ds_rm = table.repmeas('y', 'x', 'rm', ds=ds)
        >>> print(ds_rm)
        rm   a   b
        ----------
        1    1   4
        2    2   5
        3    3   6
        >>> ds_rm['difference'] = ds_rm.eval("b - a")
        >>> print(ds_rm)
        rm   a   b   difference
        -----------------------
        1    1   4   3
        2    2   5   3
        3    3   6   3

    """
    ct = Celltable(y, x, match, sub, ds=ds)
    if not ct.all_within:
        raise ValueError("Incomplete data")

    out = Dataset()
    x_ = ct.groups[ct.x.cells[0]]
    if isinstance(x_, Interaction):
        for f in x_.base:
            out.add(f)
    else:
        out[ct.match.name] = x_

    for cell in ct.x.cells:
        if len(ct.data[cell]):  # for models with empty cells
            key = as_legal_dataset_key(cellname(cell, '_'))
            out[key] = ct.data[cell]

    # Transfer other variables in ds that are compatible with the rm-structure
    if ds is not None:
        out.update(ct._align_ds(ds, True, out.keys(), isuv))

    return out



from eelbrain._text import enumeration


def test_text():
    "Test report text functions"
    assert enumeration(['a', 'b', 'c']) == 'a, b and c'
    print(enumeration({'a', 'b', 'c'}))
    assert enumeration(['a', 2]) == 'a and 2'



"""Some basic statistical tests for univariate data."""
__test__ = False
from ._utils import _deprecated_alias

from ._stats.glm import ANOVA
from ._stats.test import (
    Correlation,
    TTestOneSample, TTestIndependent, TTestRelated, MannWhitneyU, WilcoxonSignedRank,
    pairwise, ttest, correlations,
    bootstrap_pairwise, lilliefors,
    pairwise_correlations,
)

# backwards compatibility
TTest1Samp = _deprecated_alias('TTest1Samp', TTestOneSample, '0.34')
TTestInd = _deprecated_alias('TTestInd', TTestIndependent, '0.34')
TTestRel = _deprecated_alias('TTestRel', TTestRelated, '0.34')
anova = _deprecated_alias('anova', ANOVA, '0.34')



"""Tools for importing data through mne.

.. autosummary::
   :toctree: generated

   events
   epochs
   mne_epochs
   add_epochs
   add_mne_epochs
   epochs_ndvar
   evoked_ndvar
   sensor_dim
   stc_ndvar
   forward_operator
   inverse_operator
   DatasetSTCLoader


.. currentmodule:: eelbrain

Managing events with a :class:`Dataset`
---------------------------------------

To load events as :class:`Dataset`::

    >>> ds = load.fiff.events(raw_file_path)

By default, the :class:`Dataset` contains a variable called ``"trigger"``
with trigger values, and a variable called ``"i_start"`` with the indices of
the events::

    >>> print(ds[:10])
    trigger   i_start
    -----------------
    2         27977
    3         28345
    1         28771
    4         29219
    2         29652
    3         30025
    1         30450
    4         30839
    2         31240
    3         31665

These events can be modified in ``ds`` (adding event labels as :class:`Factor`,
discarding events, modifying ``i_start``, , ...) before being used to load data
epochs.

Epochs can be loaded as :class:`NDVar` with :func:`load.fiff.epochs`. Epochs
will be loaded based only on the ``"i_start"`` variable, so any modification
to this variable will affect the epochs that are loaded.::

    >>> ds['epochs'] = load.fiff.epochs(ds)

Epochs can also be loaded as mne-python :class:`mne.Epochs` object::

    >>> mne_epochs = load.fiff.mne_epochs(ds)


Using Threshold Rejection
-------------------------

In case threshold rejection is used, the number of the epochs returned by
``load.fiff.epochs(ds, reject=reject_options)`` might not be the same as the
number of events in ``ds`` (whenever epochs are rejected). For those cases,
:func:`load.fiff.add_epochs`` will automatically resize the :class:`Dataset`::

    >>> epoch_ds = load.fiff.add_epochs(ds, -0.1, 0.6, reject=reject_options)

The returned ``epoch_ds`` will contain the epochs as NDVar as ``ds['meg']``.
If no epochs got rejected during loading, the length of ``epoch_ds`` is
identical with the input ``ds``. If epochs were rejected, ``epoch_ds`` is a
shorter copy of the original ``ds``.

:class:`mne.Epochs` can be added to ``ds`` in the same fashion with::

    >>> ds = load.fiff.add_mne_epochs(ds, -0.1, 0.6, reject=reject_options)


Separate events files
---------------------

If events are stored separately form the raw files, they can be loaded in
:func:`load.fiff.events` by supplying the path to the events file as
``events`` parameter::

    >>> ds = load.fiff.events(raw_file_path, events=events_file_path)


Loading source estimates into a Dataset
---------------------------------------

Previously exported stc files can be loaded into a :class:`Dataset` with the
:class:`~load.fiff.DatasetSTCLoader` class. The stcs must reside in
subdirectories named by condition. Supply the path to the data, and the
constructor will detect the factors' levels from the names of the condition
directories. Call :meth:`~load.fiff.DatasetSTCLoader.set_factor_names` to
indicate the names of the experimental conditions, and finally load the data
with :meth:`~load.fiff.DatasetSTCLoader.make_dataset`.

    >>> loader = load.fiff.DatasetSTCLoader("path/to/exported/stcs")
    >>> loader.set_factor_names(["factor1", "factor2"])
    >>> ds = loader.make_dataset(subjects_dir="mri/")

"""
from .._io.fiff import (
    add_epochs, add_mne_epochs, epochs, epochs_ndvar, events, evoked_ndvar,
    forward_operator, inverse_operator, mne_epochs, mne_raw, raw_ndvar,
    sensor_dim, stc_ndvar, variable_length_mne_epochs,
)

from .._io.stc_dataset import DatasetSTCLoader



# Author: Christian Brodbeck <christianbrodbeck@nyu.edu>
from itertools import chain
import logging
from typing import Any, Dict, Optional

from .._exceptions import DefinitionError
from .._text import enumeration, plural
from .._utils.parse import find_variables


class CodeBase:
    _sep = ' '

    def __init__(
            self,
            string: str,
            code_string: str = None,
    ):
        self.string = string
        if code_string is None:
            code_string = string
        self._items = code_string.split(self._sep)
        self._i = -1
        self.lookahead_1 = self.lookahead()

    @classmethod
    def coerce(cls, obj):
        if isinstance(obj, cls):
            return obj
        else:
            return cls(obj)

    def __repr__(self):
        return f"{self.__class__.__name__}({self.string!r})"

    def next(self, lookahead=0):
        self._i += 1
        self.lookahead_1 = self.lookahead()
        return self.lookahead(lookahead)

    def __getitem__(self, item):
        return self._items[item]

    def __iter__(self):
        i_max = len(self._items) - 1
        while self._i < i_max:
            self._i += 1
            yield self._items[self._i]

    def assert_done(self):
        if self._i < len(self._items) - 1:
            raise self.error("Code not processed completely")

    def error(self, message=None, i=None):
        if i is None:
            i = self._i
        elif not isinstance(i, int):
            raise RuntimeError(f"i={i!r}")

        if i < 0:
            code_str = self.string
        else:
            n = len(self._items)
            if i >= n:
                code_str = f"{self._sep.join(self._items)} >unexpected end<"
            else:
                code_str = ''
                if i:
                    code_str += self._sep.join(chain(self._items[:i], [''])) + ' '
                code_str += f"> {self._items[i]} <"
                if i < n - 1:
                    code_str += ' ' + self._sep.join(chain([''], self._items[i + 1:]))
        return CodeError(code_str, message)

    def lookahead(self, offset=1):
        "Retrieve item without advancing iterator"
        i = self._i + offset
        if 0 <= i < len(self._items):
            return self._items[i]
        else:
            return ''


class FieldCode(CodeBase):
    pass


class CodeError(Exception):

    def __init__(self, code: str, message: str = None):
        if message:
            error_str = f"{code}: {message}"
        else:
            error_str = code
        Exception.__init__(self, error_str)


class Definition:
    DICT_ATTRS = None

    def as_dict(self):
        return {k: getattr(self, k) for k in self.DICT_ATTRS}

    def __eq__(self, other):
        if isinstance(other, dict):
            return self.as_dict() == other
        elif isinstance(other, Definition):
            return self.as_dict() == other.as_dict()
        else:
            return False


def name_ok(key: str, allow_empty: bool) -> bool:
    if not key and not allow_empty:
        return False
    try:
        return all(c not in key for c in ' ')
    except TypeError:
        return False


def check_names(keys, attribute, allow_empty: bool):
    invalid = [key for key in keys if not name_ok(key, allow_empty)]
    if invalid:
        raise DefinitionError(f"Invalid {plural('name', len(invalid))} for {attribute}: {enumeration(invalid)}")


def compound(items):
    out = ''
    for item in items:
        if item == '*':
            if not out.endswith('*'):
                out += '*'
        elif item:
            if out and not out.endswith('*'):
                out += ' '
            out += item
    return out


def dict_change(
        old: Dict[str, Any],
        new: Dict[str, Any],
):
    "Readable representation of dict change"
    lines = []
    keys = set(new)
    keys.update(old)
    for key in sorted(keys):
        if key not in new:
            lines.append("%s: %r -> key removed" % (key, old[key]))
        elif key not in old:
            lines.append("%s: new key -> %r" % (key, new[key]))
        elif new[key] != old[key]:
            lines.append("%s: %r -> %r" % (key, old[key], new[key]))
    return lines


def log_dict_change(
        log: logging.Logger,
        kind: str,
        name: str,
        old: Optional[Dict[str, Any]],
        new: Optional[Dict[str, Any]],
):
    if new is None:
        log.warning("  %s %s removed", kind, name)
    elif old is None:
        log.info("  %s %s added", kind, name)
    else:
        log.warning("  %s %s changed:", kind, name)
        for line in dict_change(old, new):
            log.warning("    %s", line)


def log_list_change(log, kind, name, old, new):
    log.warning("  %s %s changed:", kind, name)
    removed = tuple(v for v in old if v not in new)
    if removed:
        log.warning("    Members removed: %s", ', '.join(map(str, removed)))
    added = tuple(v for v in new if v not in old)
    if added:
        log.warning("    Members added: %s", ', '.join(map(str, added)))


def find_epoch_vars(params):
    "Find variables used in a primary epoch definition"
    out = set()
    if params.get('sel'):
        out.update(find_variables(params['sel']))
    if 'trigger_shift' in params and isinstance(params['trigger_shift'], str):
        out.add(params['trigger_shift'])
    if 'post_baseline_trigger_shift' in params:
        out.add(params['post_baseline_trigger_shift'])
    return out


def find_epochs_vars(epochs):
    "Find variables used in all epochs"
    todo = list(epochs)
    out = {}
    while todo:
        for e in tuple(todo):
            p = epochs[e]
            if 'sel_epoch' in p:
                if p['sel_epoch'] in out:
                    out[e] = find_epoch_vars(p)
                    out[e].update(out[p['sel_epoch']])
                    todo.remove(e)
            elif 'sub_epochs' in p:
                if all(se in out for se in p['sub_epochs']):
                    out[e] = find_epoch_vars(p)
                    for se in p['sub_epochs']:
                        out[e].update(out[se])
                    todo.remove(e)
            elif 'collect' in p:
                if all(se in out for se in p['collect']):
                    out[e] = find_epoch_vars(p)
                    for se in p['collect']:
                        out[e].update(out[se])
                    todo.remove(e)
            else:
                out[e] = find_epoch_vars(p)
                todo.remove(e)
    return out


def find_dependent_epochs(epoch, epochs):
    "Find all epochs whose definition depends on epoch"
    todo = set(epochs).difference(epoch)
    out = [epoch]
    while todo:
        last_len = len(todo)
        for e in tuple(todo):
            p = epochs[e]
            if 'sel_epoch' in p:
                if p['sel_epoch'] in out:
                    out.append(e)
                    todo.remove(e)
            elif 'sub_epochs' in p:
                if any(se in out for se in p['sub_epochs']):
                    out.append(e)
                    todo.remove(e)
            elif 'collect' in p:
                if any(se in out for se in p['collect']):
                    out.append(e)
                    todo.remove(e)
            else:
                todo.remove(e)
        if len(todo) == last_len:
            break
    return out[1:]


def typed_arg(arg, type_, secondary_type=None):
    if secondary_type is not None and isinstance(arg, secondary_type):
        return arg
    elif arg is None:
        return None
    else:
        return type_(arg)


def tuple_arg(arg, item_type=str):
    if isinstance(arg, item_type):
        return arg,
    elif isinstance(arg, list):
        out = tuple(arg)
    elif isinstance(arg, tuple):
        out = arg
    else:
        raise TypeError(f"{arg}: tuple of {item_type.__name__} required")
    if not all(isinstance(item, item_type) for item in out):
        raise TypeError(f"{arg}: tuple of {item_type.__name__} required")
    return out



# Author: Christian Brodbeck <christianbrodbeck@nyu.edu>
"""Helper functions for saving data in various formats."""

from ._besa import meg160_triggers, besa_evt
from .._io.pickle import pickle
from ._txt import txt
from .._io.pyarrow_context import save_arrow as arrow
from .._io.wav import save_wav as wav



# Author: Christian Brodbeck <christianbrodbeck@nyu.edu>
"""Plots for color/style legends"""
from collections.abc import Iterator
from itertools import chain
from typing import Any, Dict, Literal, Sequence, Union

import numpy as np
import matplotlib as mpl
from matplotlib.colors import LinearSegmentedColormap, Colormap, Normalize, to_rgb
from matplotlib.colorbar import ColorbarBase
from matplotlib.ticker import FixedFormatter, MaxNLocator

from .._data_obj import CellArg, cellname
from .._utils import IS_WINDOWS
from ._base import EelFigure, Layout, AxisScale, CMapArg, ColorArg, fix_vlim_for_cmap


POINT_SIZE = 0.0138889  # 1 point in inches
LEGEND_SIZE = 1.2  # times font.size


class ColorGrid(EelFigure):
    """Plot colors for a two-way design in a grid

    Parameters
    ----------
    row_cells
        Cells contained in the rows.
    column_cells
        Cells contained in the columns.
    colors
        Colors for cells.
    size
        Size (width and height) of the color squares (the default is to
        scale them to fit the font size).
    column_label_position : 'top' | 'bottom' | 'none'
        Where to place the column labels (default is 'top').
    row_first
        Whether the row cell precedes the column cell in color keys. By
        default this is inferred from the existing keys.
    labels
        Condition labels that are used instead of the keys in ``row_cells`` and
        ``column_cells``.
    shape : 'box' | 'line'
        Shape for color samples (default 'box').
    ...
        Also accepts :ref:`general-layout-parameters`.

    Attributes
    ----------
    column_labels : list of :class:`matplotlib.text.Text`
        Column labels.
    row_labels : list of :class:`matplotlib.text.Text`
        Row labels.
    """
    def __init__(
            self,
            row_cells: Sequence[str],
            column_cells: Sequence[str],
            colors: Dict[CellArg, Any],
            size: float = None,
            column_label_position: str = 'top',
            row_first: bool = None,
            labels: dict = None,
            shape: str = 'box',
            **kwargs):
        row_cells = list(row_cells)
        column_cells = list(column_cells)
        if row_first is None:
            row_cell_0 = row_cells[0]
            col_cell_0 = column_cells[0]
            if (row_cell_0, col_cell_0) in colors:
                row_first = True
            elif (col_cell_0, row_cell_0) in colors:
                row_first = False
            else:
                raise KeyError(f"Neither {(row_cell_0, col_cell_0)} nor {(col_cell_0, row_cell_0)} exist as a key in colors")

        # reverse rows so we can plot upwards
        row_cells = tuple(reversed(row_cells))
        n_rows = len(row_cells)
        n_cols = len(column_cells)

        # prepare labels
        if labels:
            column_labels = [labels.get(c, c) for c in column_cells]
            row_labels = [labels.get(c, c) for c in row_cells]
        else:
            column_labels = column_cells
            row_labels = row_cells

        # default size
        chr_size = mpl.rcParams['font.size'] * POINT_SIZE
        if size is None:
            size = chr_size * LEGEND_SIZE
        w_default = size * (n_cols + 1) + chr_size * max(len(l) for l in row_labels)
        h_default = size * n_rows
        if column_label_position != 'none':
            h_default += chr_size * max(len(l) for l in column_labels)
        aspect = w_default / h_default
        layout = Layout(0, aspect, h_default, tight=False, **kwargs)
        EelFigure.__init__(self, None, layout)
        ax = self.figure.add_axes((0, 0, 1, 1), frameon=False)
        ax.set_axis_off()
        self._ax = ax

        # color patches
        for col in range(n_cols):
            for row in range(n_rows):
                if row_first:
                    cell = (row_cells[row], column_cells[col])
                else:
                    cell = (column_cells[col], row_cells[row])

                if shape == 'box':
                    patch = mpl.patches.Rectangle((col, row), 1, 1, fc=colors[cell], ec='none')
                    ax.add_patch(patch)
                elif shape == 'line':
                    y = row + 0.5
                    ax.plot([col, col + 1], [y, y], color=colors[cell])
                else:
                    raise ValueError(f"shape={shape!r}")

        # column labels
        self.column_labels = []
        if column_label_position == 'none':
            ymin = 0
            ymax = self._layout.h / size
        else:
            tilt_labels = any(len(label) > 1 for label in column_labels)
            if column_label_position == 'top':
                y = n_rows + 0.1
                va = 'bottom'
                rotation = 40 if tilt_labels else 0
                ymin = 0
                ymax = self._layout.h / size
            elif column_label_position == 'bottom':
                y = -0.1
                va = 'top'
                rotation = -40 if tilt_labels else 0
                ymax = n_rows
                ymin = n_rows - self._layout.h / size
            else:
                raise ValueError(f"column_label_position={column_label_position!r}")

            ha = 'left' if tilt_labels else 'center'
            x_offset = 0 if tilt_labels else 0.5
            y_offset = tilt_labels * size * 1.5
            for col, label in enumerate(column_labels):
                y_col = y + y_offset * (len(column_labels) - col - 1)
                h = ax.text(col + x_offset, y_col, label, va=va, ha=ha, rotation=rotation)
                self.column_labels.append(h)

        # row labels
        x = n_cols + 0.1
        self.row_labels = []
        for row, label in enumerate(row_labels):
            h = ax.text(x, row + 0.5, label, va='center', ha='left')
            self.row_labels.append(h)

        if size is not None:
            self._ax.set_xlim(0, self._layout.w / size)
            self._ax.set_ylim(ymin, ymax)

        self._show()

    def _tight(self):
        # arbitrary default with equal aspect
        self._ax.set_ylim(0, 1)
        self._ax.set_xlim(0, 1 * self._layout.w / self._layout.h)

        # draw to compute text coordinates
        self.draw()

        # find label bounding box
        xmax = 0
        ymax = 0
        for h in chain(self.column_labels, self.row_labels):
            bbox = h.get_window_extent()
            if bbox.xmax > xmax:
                xmax = bbox.xmax
                xpos = h.get_position()[0]
            if bbox.ymax > ymax:
                ymax = bbox.ymax
                ypos = h.get_position()[1]
        xmax += 2
        ymax += 2

        # transform from display coordinates -> data coordinates
        trans = self._ax.transData.inverted()
        xmax, ymax = trans.transform((xmax, ymax))

        # calculate required movement
        _, ax_xmax = self._ax.get_xlim()
        _, ax_ymax = self._ax.get_ylim()
        xtrans = ax_xmax - xmax
        ytrans = ax_ymax - ymax

        # calculate the scale factor:
        # new_coord = x * coord
        # new_coord = coord + trans
        # x = (coord + trans) / coord
        scale = (xpos + xtrans) / xpos
        scale_y = (ypos + ytrans) / ypos
        if scale_y <= scale:
            scale = scale_y

        self._ax.set_xlim(0, ax_xmax / scale)
        self._ax.set_ylim(0, ax_ymax / scale)


class ColorList(EelFigure):
    """Plot colors with labels

    Parameters
    ----------
    colors
        Colors for cells.
    cells
        Cells for which to plot colors (default is ``colors.keys()``).
    labels
        Condition labels that are used instead of the keys in ``colors``. This
        is useful if ``colors`` uses abbreviated labels, but the color legend
        should contain more intelligible labels.
    size
        Size (width and height) of the color squares (the default is to
        scale them to fit the font size).
    h : 'auto' | float
        Height of the figure in inches. If 'auto' (default), the height is
        chosen to fit all labels.
    shape : 'box' | 'line'
        Shape for color samples (default 'box').
    ...
        Also accepts :ref:`general-layout-parameters`.

    Attributes
    ----------
    labels : list of :class:`matplotlib.text.Text`
        Color labels.
    """
    def __init__(
            self,
            colors: Dict[CellArg, Any],
            cells: Sequence[CellArg] = None,
            labels: Dict[CellArg, str] = None,
            size: float = None,
            h: Union[str, float] = 'auto',
            shape: str = 'box',
            **kwargs):
        if cells is None:
            cells = tuple(colors.keys())
        elif isinstance(cells, Iterator):
            cells = tuple(cells)

        if h == 'auto':
            if size is None:
                size = mpl.rcParams['font.size'] * LEGEND_SIZE * POINT_SIZE
            h = len(cells) * size
        elif size is None:  # size = h / len(cells)
            pass
        else:
            raise NotImplementedError("specifying size and h parameters together")

        if labels is None:
            labels = {cell: cellname(cell) for cell in cells}
        elif not isinstance(labels, dict):
            raise TypeError(f"labels={labels!r}")

        layout = Layout(0, 1.5, 2, False, h=h, **kwargs)
        EelFigure.__init__(self, None, layout)

        ax = self.figure.add_axes((0, 0, 1, 1), frameon=False)
        ax.set_axis_off()

        n = len(cells)
        self.labels = []
        for i, cell in enumerate(cells):
            bottom = n - i - 1
            y = bottom + 0.5
            if shape == 'box':
                patch = mpl.patches.Rectangle((0, bottom), 1, 1, fc=colors[cell], ec='none', zorder=1)
                ax.add_patch(patch)
            elif shape == 'line':
                ax.plot([0, 1], [y, y], color=colors[cell])
            else:
                raise ValueError(f"shape={shape!r}")
            h = ax.text(1.1, y, labels.get(cell, cell), va='center', ha='left', zorder=2)
            self.labels.append(h)

        ax.set_ylim(0, n)
        ax.set_xlim(0, n * self._layout.w / self._layout.h)

        self._draw_hooks.append(self.__update_frame)
        self._ax = ax
        self._show()
        if IS_WINDOWS and self._has_frame:
            self._frame.Fit()

    def __update_frame(self):
        if self._layout.w_fixed or not self._has_frame:
            return

        # resize figure to match legend
        # (all calculation in pixels)
        fig_bb = self.figure.get_window_extent()
        x_max = max(h.get_window_extent().x1 for h in self.labels)
        w0, h0 = self._frame.GetSize()
        new_w = w0 + (x_max - fig_bb.x1) + 5
        self._frame.SetSize((new_w, h0))
        # adjust x-limits
        n = len(self.labels)
        ax_bb = self._ax.get_window_extent()
        self._ax.set_xlim(0, n * ax_bb.width / ax_bb.height)


class ColorBar(EelFigure):
    """A color-bar for a matplotlib color-map

    Parameters
    ----------
    cmap
        Name of the color-map, or a matplotlib Colormap, or LUT.
    vmin
        Lower end of the scale mapped onto cmap.
    vmax
        Upper end of the scale mapped onto cmap.
    label
        Label for the x-axis (default is the unit, or if no unit is provided
        the name of the colormap).
    label_position
        Position of the axis label. Valid values depend on orientation.
    label_rotation
        Angle of the label in degrees (For horizontal colorbars, the default is
        0; for vertical colorbars, the default is 0 for labels of 3 characters
        and shorter, and 90 for longer labels).
    clipmin
        Clip the color-bar below this value.
    clipmax
        Clip the color-bar above this value.
    orientation
        Orientation of the bar (default is horizontal).
    unit
        Unit for the axis to determine tick labels (for example, ``u'µV'`` to
        label 0.000001 as '1').
    contours
        Plot contour lines at these values.
    width
        Width of the color-bar in inches.
    ticks
        Customize tick-labels on the colormap; either a dictionary with
        tick-locations and labels, or a sequence of tick locations. To draw no
        ticks, set to ``()``.
    threshold
        Set the alpha of values below ``threshold`` to 0 (as well as for
        negative values above ``abs(threshold)``).
    ticklocation
        Where to place ticks and label.
    background : matplotlib color
        Background color (for colormaps including transparency).
    ...
        Also accepts :ref:`general-layout-parameters`.
    """
    def __init__(
            self,
            cmap: CMapArg,
            vmin: float = None,
            vmax: float = None,
            label: Union[bool, str] = True,
            label_position: Literal['left', 'right', 'top', 'bottom'] = None,
            label_rotation: float = None,
            clipmin: float = None,
            clipmax: float = None,
            orientation: Literal['horizontal', 'vertical'] = 'horizontal',
            unit: str = None,
            contours: Sequence[float] = (),
            width: float = None,
            ticks: Union[Dict[float, str], Sequence[float]] = None,
            threshold: float = None,
            ticklocation: Literal['auto', 'top', 'bottom', 'left', 'right'] = 'auto',
            background: ColorArg = 'white',
            tight: bool = True,
            h: float = None,
            w: float = None,
            **kwargs,
    ):
        # get Colormap
        if isinstance(cmap, np.ndarray):
            if threshold is not None:
                raise NotImplementedError("threshold parameter with cmap=<array>")
            if cmap.max() > 1:
                cmap = cmap / 255.
            cm = mpl.colors.ListedColormap(cmap, 'LUT')
        elif isinstance(cmap, Colormap):
            cm = cmap
        else:
            cm = mpl.cm.get_cmap(cmap)

        # prepare layout
        if orientation == 'horizontal':
            if h is None and w is None:
                h = 1
            ax_aspect = 4
        elif orientation == 'vertical':
            if h is None and w is None:
                h = 4
            ax_aspect = 0.3
        else:
            raise ValueError("orientation=%s" % repr(orientation))

        layout = Layout(1, ax_aspect, 2, tight, h=h, w=w, **kwargs)
        EelFigure.__init__(self, cm.name, layout)
        ax = self.axes[0]

        # translate between axes and data coordinates
        if isinstance(vmin, Normalize):
            norm = vmin
        else:
            vmin, vmax = fix_vlim_for_cmap(vmin, vmax, cm)
            norm = Normalize(vmin, vmax)

        if isinstance(unit, AxisScale):
            scale = unit
        else:
            scale = AxisScale(unit or 1, label)

        # value ticks
        if ticks is False:
            tick_locs = ()
            formatter = scale.formatter
        elif isinstance(ticks, dict):
            tick_locs = sorted(ticks)
            formatter = FixedFormatter([ticks[t] for t in tick_locs])
        else:
            if ticks is None:
                tick_locs = MaxNLocator(4)
            else:
                tick_locs = ticks
            formatter = scale.formatter

        if orientation == 'horizontal':
            axis = ax.xaxis
            contour_func = ax.axhline
        else:
            axis = ax.yaxis
            contour_func = ax.axvline

        if label is True:
            label = scale.label or cm.name
        if not label:
            label = ''

        # show only part of the colorbar
        if clipmin is not None or clipmax is not None:
            boundaries = norm.inverse(np.linspace(0, 1, cm.N + 1))
            if clipmin is None:
                start = None
            else:
                start = np.digitize(clipmin, boundaries)
                boundaries[start] = clipmin
            if clipmax is None:
                stop = None
            else:
                stop = np.digitize(clipmax, boundaries) + 1
                boundaries[stop-1] = clipmax
            boundaries = boundaries[start:stop]
        else:
            boundaries = None

        colorbar = ColorbarBase(ax, cmap=cm, norm=norm, boundaries=boundaries, orientation=orientation, ticklocation=ticklocation, ticks=tick_locs, label=label, format=formatter)

        # label position/rotation
        if label_position is not None:
            axis.set_label_position(label_position)
        if label_rotation is not None:
            axis.label.set_rotation(label_rotation)
            if orientation == 'vertical':
                if (label_rotation + 10) % 360 < 20:
                    axis.label.set_va('center')
        elif orientation == 'vertical' and len(label) <= 3:
            axis.label.set_rotation(0)
            axis.label.set_va('center')

        self._contours = [contour_func(c, c='k') for c in contours]
        self._draw_hooks.append(self.__fix_alpha)
        self._draw_hooks.append(self.__update_bar_tickness)

        self._background = background
        self._colorbar = colorbar
        self._orientation = orientation
        self._width = width
        self._show()

    def __fix_alpha(self):
        # fix cmaps with alpha https://stackoverflow.com/q/15003353/166700
        if self._background is not False:
            lut = self._colorbar.solids.get_facecolor()
            bg_color = to_rgb(self._background)
            lut[:, :3] *= lut[:, 3:]
            lut[:, :3] += (1 - lut[:, 3:]) * bg_color
            lut[:, 3] = 1.
            self._colorbar.solids.set_facecolor(lut)
            return True

    def _tight(self):
        # make sure ticklabels have space
        ax = self.axes[0]
        if self._orientation == 'vertical' and not self._layout.w_fixed:
            self.draw()
            labels = ax.get_yticklabels()
            # wmax = max(l.get_window_extent().width for l in labels)
            x0 = min(l.get_window_extent().x0 for l in labels)
            if x0 < 0:
                w, h = self.figure.get_size_inches()
                w -= (x0 / self._layout.dpi)
                self.figure.set_size_inches(w, h, forward=True)
        super(ColorBar, self)._tight()

    def __update_bar_tickness(self):
        # Override to keep bar thickness
        if not self._width:
            return
        ax = self.axes[0]
        x = (self._width, self._width)
        x = self.figure.dpi_scale_trans.transform(x)
        x = self.figure.transFigure.inverted().transform(x)
        pos = ax.get_position()
        xmin, ymin, width, height = pos.xmin, pos.ymin, pos.width, pos.height
        if self._orientation == 'vertical':
            if self._layout._margins_arg and 'right' in self._layout._margins_arg:
                xmin += width - x[0]
            width = x[0]
        else:
            if self._layout._margins_arg and 'top' in self._layout._margins_arg:
                ymin += height - x[1]
            height = x[1]
        ax.set_position((xmin, ymin, width, height))
        return True



"""Predictors for reverse correlation"""
# Author: Christian Brodbeck <christianbrodbeck@nyu.edu>
from itertools import repeat

import numpy as np

from .._data_obj import NDVar, Case, UTS, asndvar, asarray


def epoch_impulse_predictor(shape, value=1, latency=0, name=None, ds=None):
    """Time series with one impulse for each of ``n`` epochs

    Parameters
    ----------
    shape : NDVar | (int, UTS) | str
        Shape of the output. Can be specified as the :class:`NDVar` with the
        data to predict, or an ``(n_cases, time_dimension)`` tuple.
    value : scalar | sequence | str
        Scalar or length ``n`` sequence of scalars specifying the value of each
        impulse (default 1).
    latency : scalar | sequence | str
        Scalar or length ``n`` sequence of scalars specifying the latency of
        each impulse (default 0).
    name : str
        Name for the output :class:`NDVar`.
    ds : Dataset
        If specified, input items (``shape``, ``value`` and ``latency``) can be
        strings to be evaluated in ``ds``.

    Examples
    --------
    See :ref:`exa-impulse` example.
    """
    if isinstance(shape, str):
        shape = asndvar(shape, ds=ds)
    if isinstance(value, str):
        value = asarray(value, ds=ds)
    if isinstance(latency, str):
        latency = asarray(latency, ds=ds)

    if isinstance(shape, NDVar):
        if not shape.has_case:
            raise ValueError(f'shape={shape!r}: has no case dimension')
        n = len(shape)
        time = shape.get_dim('time')
    else:
        n, time = shape
        if not isinstance(time, UTS):
            raise TypeError(f'shape={shape!r}: second item needs to be UTS instance')

    x = np.zeros((n, len(time)))
    t_index = time._array_index(latency)
    if np.isscalar(latency):
        x[:, t_index] = value
    else:
        x[np.arange(n), t_index] = value
    return NDVar(x, (Case, time), name)


def event_impulse_predictor(shape, time='time', value=1, latency=0, name=None, ds=None):
    """Time series with multiple impulses

    Parameters
    ----------
    shape : NDVar | UTS
        Shape of the output. Can be specified as the :class:`NDVar` with the
        data to predict, or an ``(n_cases, time_dimension)`` tuple.
    time : sequence of scalar
        Time points at which impulses occur.
    value : scalar | sequence
        Magnitude of each impulse (default 1).
    latency : scalar | sequence
        Latency of each impulse relative to ``time`` (default 0).
    name : str
        Name for the output :class:`NDVar`.
    ds : Dataset
        If specified, input items (``time``, ``value`` and ``latency``) can be
        strings to be evaluated in ``ds``.
    """
    if isinstance(shape, NDVar):
        uts = shape.get_dim('time')
    elif isinstance(shape, UTS):
        uts = shape
    else:
        raise TypeError(f'shape={shape!r}')

    time, n = asarray(time, ds=ds, return_n=True)

    if isinstance(value, str) or not np.isscalar(value):
        value = asarray(value, ds=ds, n=n)
    else:
        value = repeat(value)

    if isinstance(latency, str) or not np.isscalar(latency):
        latency = asarray(latency, ds=ds, n=n)
    else:
        latency = repeat(latency)

    out = NDVar(np.zeros(len(uts)), uts, name)
    for t, l, v in zip(time, latency, value):
        out[t + l] = v
    return out



# Author: Christian Brodbeck <christianbrodbeck@nyu.edu>
import numpy as np
from numpy.testing import assert_equal, assert_array_equal
import pytest

from eelbrain import datasets, Celltable, testnd
from eelbrain._stats import t_contrast
from eelbrain._stats.stats import t_1samp
from eelbrain._stats.t_contrast import TContrastSpec


def test_t_contrast_parsing():
    "Test parsing of t-contrast expressions"
    y = np.arange(9.).reshape((3, 3))
    indexes = {'a': 0, 'b': 1, 'c': 2}

    contrast = "sum(a>c, b>c)"
    contrast_ = t_contrast.parse(contrast)
    assert contrast_ == ('afunc', np.sum, (('comp', 'a', 'c'), ('comp', 'b', 'c')))
    _, cells = t_contrast._t_contrast_rel_properties(contrast_)
    pc, mc = t_contrast._t_contrast_rel_expand_cells(cells, ('a', 'b', 'c'))
    data = t_contrast._t_contrast_rel_data(y, indexes, pc, mc)
    assert_equal(data['a'], np.arange(0., 3.))
    assert_equal(data['b'], np.arange(3., 6.))
    assert_equal(data['c'], np.arange(6., 9.))

    contrast = "sum(a>*, b>*)"
    contrast_ = t_contrast.parse(contrast)
    assert contrast_ == ('afunc', np.sum, (('comp', 'a', '*'), ('comp', 'b', '*')))
    _, cells = t_contrast._t_contrast_rel_properties(contrast_)
    pc, mc = t_contrast._t_contrast_rel_expand_cells(cells, ('a', 'b', 'c'))
    data = t_contrast._t_contrast_rel_data(y, indexes, pc, mc)
    assert_equal(data['a'], np.arange(0., 3.))
    assert_equal(data['b'], np.arange(3., 6.))
    assert_equal(data['*'], y.mean(0))

    with pytest.raises(ValueError):
        t_contrast._t_contrast_rel_expand_cells(cells, ('a|c', 'b|c', 'c|c'))

    # test finding cells
    all_cells = (('fondue pot', 'brie'), ('fondue mix', 'brie'),
                 ('fondue pot', 'edam'), ('raclette', 'edam'))
    cells = (('* pot', '*'), ('fondue *', 'brie'))
    pc, mc = t_contrast._t_contrast_rel_expand_cells(cells, all_cells)
    assert pc == set(all_cells[:3])
    assert mc == {
        ('* pot', '*'): (('fondue pot', 'brie'), ('fondue pot', 'edam')),
        ('fondue *', 'brie'): (('fondue pot', 'brie'), ('fondue mix', 'brie'))}


def test_t_contrasts():
    "Test computation of various t-contrasts"
    ds = datasets.get_uts()
    ct = Celltable('uts', 'A % B', 'rm', ds=ds)
    y = ct.y.x
    out = np.empty(y.shape[1:])
    perm = np.arange(ds.n_cases)
    a1b1 = ct.data['a1', 'b1'].x
    a1b0 = ct.data['a1', 'b0'].x
    a0b1 = ct.data['a0', 'b1'].x
    a0b0 = ct.data['a0', 'b0'].x

    # simple t-test
    tgt = t_1samp(a1b0 - a0b0)
    c = TContrastSpec("a1|b0 > a0|b0", ct.cells, ct.data_indexes)
    assert_equal(c.map(y), tgt)
    assert_equal(c(y, out, perm), tgt)
    out.fill(0)
    assert_equal(c(y, out, perm), tgt)
    c = TContrastSpec("(a1|b0 > a0|b0)", ct.cells, ct.data_indexes)
    assert_equal(c.map(y), tgt)
    assert_equal(c(y, out, perm), tgt)
    out.fill(0)
    assert_equal(c(y, out, perm), tgt)

    # intersection
    tgt = np.min((t_1samp(a1b0 - a0b0), t_1samp(a1b1 - a0b1)), 0)
    c = TContrastSpec("min(a1|b0 > a0|b0, a1|b1 > a0|b1)", ct.cells, ct.data_indexes)
    assert_equal(c.map(y), tgt)
    assert_equal(c(y, out, perm), tgt)
    out.fill(0)
    assert_equal(c(y, out, perm), tgt)

    # unary function
    tgt = np.abs(t_1samp(a1b0 - a0b0))
    c = TContrastSpec("abs(a1|b0 > a0|b0)", ct.cells, ct.data_indexes)
    assert_equal(c.map(y), tgt)
    assert_equal(c(y, out, perm), tgt)
    out.fill(0)
    assert_equal(c(y, out, perm), tgt)

    # "VMod"
    tgt = t_1samp(a1b0 - a0b0) - np.abs(t_1samp(a1b1 - a0b1))
    c = TContrastSpec("(a1|b0 > a0|b0) - abs(a1|b1 > a0|b1)", ct.cells,
                      ct.data_indexes)
    assert_equal(c.map(y), tgt)

    c = TContrastSpec("subtract(a1|b0 > a0|b0, abs(a1|b1 > a0|b1))", ct.cells,
                      ct.data_indexes)
    assert_equal(c.map(y), tgt)
    assert_equal(c(y, out, perm), tgt)
    out.fill(0)
    assert_equal(c(y, out, perm), tgt)


def test_t_contrast_testnd():
    ds = datasets.get_uts()

    # binary function
    res = testnd.TContrastRelated('uts', 'A', "a1>a0 - a0>a1", 'rm', ds=ds, tmin=4, samples=10)
    assert_equal(res.find_clusters()['p'], np.array([1, 1, 0.9, 0, 0.2, 1, 1, 0]))
    res_t = testnd.TTestRelated('uts', 'A', 'a1', 'a0', match='rm', ds=ds, tmin=2, samples=10)
    assert_array_equal(res.t.x, res_t.t.x * 2)
    assert_array_equal(res.clusters['tstart'], res_t.clusters['tstart'])
    assert_array_equal(res.clusters['tstop'], res_t.clusters['tstop'])
    assert_array_equal(res.clusters['v'], res_t.clusters['v'] * 2)



"""Deploy UI elements depending on the current environment."""
import os

USE_WX = True

_progress_monitors = []


def get_ui():
    global USE_WX

    if USE_WX:
        try:
            from . import wx_ui
            from ..._wxgui import get_app
            get_app()
            return wx_ui
        except ImportError:
            USE_WX = False

    from . import tk_ui
    return tk_ui


def ask_saveas(title="Save File", message="Please Pick a File Name",
               filetypes=(("Pickled Python Objects (*.pickled)", '*.pickled'),),
               defaultDir=None, defaultFile=False):
    """Display a save-as dialog

    Parameters
    ----------
    title : str
        Title of the dialog.
    message : str
        Message in the dialog.
    filetypes : sequence of tuples
        Sequence of (label, pattern) tuples.
    defaultDir : None | str
        Default directory to save at.
    defaultFile : None | str
        Default file name.

    Returns
    -------
    result : False | str
        The path as str if the user selects a dialog, otherwise ``False``.
    """
    result = get_ui().ask_saveas(title, message, filetypes, defaultDir,
                                 defaultFile)
    return result


def ask_dir(title="Select Folder",
            message="Please Pick a Folder",
            must_exist=True):
    return get_ui().ask_dir(title, message, must_exist)


def ask_file(title="Pick File",
             message="Please Pick a File",
             filetypes=[("All files", '*')],
             directory='',
             mult=False):
    """
    Ask for an existing file.

    Parameters
    ----------
    title, message : str
        Title and message for the dialog.
    filetypes : sequence of tuples
        Sequence of (label, pattern) tuples.
    directory : None | str
        Path to initial directory.
    mult : bool
        Allow selecting multiple files.

    Returns
    -------
    paths : False | str | list
        If the user cancels: False. Otherwise, if mult=False a single path, and
        if mult=True a list of paths.
    """
    return get_ui().ask_file(title, message, filetypes, directory, mult)


def ask(title="Overwrite File?",
        message="Duplicate filename. Do you want to overwrite?",
        cancel=False,
        default=True,  # True=YES, False=NO, None=Nothing
        ):
    return get_ui().ask(title, message, cancel, default)


def ask_color(default=(0, 0, 0)):
    return get_ui().ask_color(default)


def ask_str(msg, title, default=''):
    return get_ui().ask_str(msg, title, default)


def kill_progress_monitors():
    while len(_progress_monitors) > 0:
        p = _progress_monitors.pop()
        p.terminate()


def message(title, message="", icon='i'):
    return get_ui().message(title, message, icon)


def copy_file(path):
    return get_ui().copy_file(path)


def copy_text(text):
    return get_ui().copy_text(text)


def test_targetpath(path, cancel=True):
    """Test whether ``path`` is okay to write to

    If the directory does not exist, the user is asked whether it should be
    created.

    Parameters
    ----------
    path : str
        Path to test.
    cancel : bool
        Add a cancel button. If clicked, a KeyboardInterrupt Exception is
        raised.

    Returns
    -------
    success : bool
        True if path is a valid path to write to, False otherwise.
    """
    if not path:
        return False

    dirname = os.path.abspath(os.path.dirname(path))
    if not os.path.exists(dirname):
        msg = ("The directory %r does not exist. Should it be created?" % dirname)
        answer = ask("Create Directory?", msg, cancel=cancel)
        if answer:
            os.makedirs(dirname)
        elif answer is None:  # cancel
            err = ("User canceled because the directory %r does not exist"
                   % dirname)
            raise KeyboardInterrupt(err)

    return os.path.exists(dirname)



# Author: Christian Brodbeck <christianbrodbeck@nyu.edu>
from typing import List
from warnings import catch_warnings, filterwarnings

import numpy as np
from scipy.linalg import norm
from scipy.stats import spearmanr, SpearmanRConstantInputWarning

from ._boosting_opt import l1, l2
from .shared import RevCorrData


class Evaluator:
    vector = False
    attr = NotImplemented
    name = NotImplemented
    meas = None

    def __init__(
            self,
            data: RevCorrData,
            segments: List[np.ndarray] = None,  # evaluations for different test-segments
    ):
        n = len(data.y)
        if self.vector:
            if not data.vector_dim:
                raise ValueError(f"{self.__class__.__name__}: Vector evaluator for non-vector data")
            n //= len(data.vector_dim)
        self.data = data
        self.segments = segments
        self.xs = [np.empty(n, np.float64) for _ in segments]

    @classmethod
    def _crop_y(cls, segments, *ys):
        "For error functions that require one contiguous segment"
        if len(segments) == 1:
            istart, istop = segments[0]
            out = [y[..., istart: istop] for y in ys]
        else:
            # OPT: avoid new array creation
            out = [np.concatenate([y[..., i_start:i_stop] for i_start, i_stop in segments], axis=-1) for y in ys]
        return out[0] if len(out) == 1 else out


    def add_y(
            self,
            i: int,  # y index (row in data.y)
            y: np.ndarray,  # actual data
            y_pred: np.ndarray,  # data predicted by model
    ):
        raise NotImplementedError

    def __repr__(self):
        return f"<{self.__class__.__name__} evaluator>"

    def get(self, i_test: int = -1):
        return self.data.package_value(self.xs[i_test + 1], self.name, meas=self.meas)


class L1(Evaluator):
    attr = 'residual'
    name = 'L1 residuals'

    def add_y(
            self,
            i: int,  # y index (row in data.y)
            y: np.ndarray,  # actual data
            y_pred: np.ndarray,  # data predicted by model
    ):
        err = y - y_pred
        for x, segments in zip(self.xs, self.segments):
            x[i] = l1(err, segments)


class L2(Evaluator):
    attr = 'residual'
    name = 'L2 residuals'

    def add_y(
            self,
            i: int,  # y index (row in data.y)
            y: np.ndarray,  # actual data
            y_pred: np.ndarray,  # data predicted by model
    ):
        err = y - y_pred
        for x, segments in zip(self.xs, self.segments):
            x[i] = l2(err, segments)


class Correlation(Evaluator):
    attr = 'r'
    name = 'Correlation'
    meas = 'r'

    def add_y(
            self,
            i: int,  # y index (row in data.y)
            y: np.ndarray,  # actual data
            y_pred: np.ndarray,  # data predicted by model
    ):
        for x, segments in zip(self.xs, self.segments):
            y_i, y_pred_i = self._crop_y(segments, y, y_pred)
            with catch_warnings():
                filterwarnings('ignore', "invalid value encountered", RuntimeWarning)
                r = np.corrcoef(y_i, y_pred_i)[0, 1]
            x[i] = 0 if np.isnan(r) else r


class RankCorrelation(Evaluator):
    attr = 'r_rank'
    name = 'Rank correlation'
    meas = 'r'

    def add_y(
            self,
            i: int,  # y index (row in data.y)
            y: np.ndarray,  # actual data
            y_pred: np.ndarray,  # data predicted by model
    ):
        for x, segments in zip(self.xs, self.segments):
            y_i, y_pred_i = self._crop_y(segments, y, y_pred)
            with catch_warnings():
                filterwarnings('ignore', "invalid value encountered", RuntimeWarning)
                filterwarnings('ignore', category=SpearmanRConstantInputWarning)
                r = spearmanr(y_i, y_pred_i)[0]
            x[i] = 0 if np.isnan(r) else r


class VectorL1(Evaluator):
    vector = True
    attr = 'residual'
    name = 'Vector l1 residuals'

    def add_y(
            self,
            i: int,  # y index (row in data.y)
            y: np.ndarray,  # actual data
            y_pred: np.ndarray,  # data predicted by model
    ):
        err = norm(y - y_pred, axis=0)
        for x, segments in zip(self.xs, self.segments):
            err_i = self._crop_y(segments, err)
            x[i] = err_i.sum(-1)


class VectorL2(Evaluator):
    vector = True
    attr = 'residual'
    name = 'Vector l2 residuals'

    def add_y(
            self,
            i: int,  # y index (row in data.y)
            y: np.ndarray,  # actual data
            y_pred: np.ndarray,  # data predicted by model
    ):
        err = y - y_pred
        err **= 2
        err = err.sum(0)
        for x, segments in zip(self.xs, self.segments):
            err_i = self._crop_y(segments, err)
            x[i] = err_i.sum()


class VectorCorrelation(Evaluator):
    vector = True
    attr = 'r'
    name = 'Vector correlation'
    meas = 'r'

    def add_y(
            self,
            i: int,  # y index (row in data.y)
            y: np.ndarray,  # actual data
            y_pred: np.ndarray,  # data predicted by model
    ):
        for x, segments in zip(self.xs, self.segments):
            y, y_pred = self._crop_y(segments, y, y_pred)
            x[i] = self._r(y, y_pred)

    def _r(self, y, y_pred):
        # shape (space, time)
        y_norm = norm(y, axis=0)
        y_pred_norm = norm(y_pred, axis=0)
        # rescale y_pred
        y_pred_scale = (y_pred_norm ** 2).mean() ** 0.5
        if y_pred_scale == 0:
            return 0
        y_pred_l2 = y_pred / y_pred_scale
        # rescale y
        y_scale = (y_norm ** 2).mean() ** 0.5
        if y_scale == 0:
            return 0
        elif self.data.scale_data == 'l2':
            y_l2 = y
        else:
            y_l2 = y / y_scale
        # l2 correlation
        return np.multiply(y_l2, y_pred_l2, out=y_pred_l2).sum(0).mean(0)


class VectorCorrelationL1(VectorCorrelation):
    vector = True
    attr = 'r_l1'
    name = 'Vector l1-correlation'
    meas = 'r'

    def _r(self, y, y_pred):
        # shape (space, time)
        y_norm = norm(y, axis=0)
        y_pred_norm = norm(y_pred, axis=0)
        # rescale y_pred
        y_pred_scale = y_pred_norm.mean()
        if y_pred_scale == 0:
            return 0
        y_pred_l1 = y_pred / y_pred_scale
        # rescale y
        y_scale = y_norm.mean()
        if y_scale == 0:
            return 0
        elif self.data.scale_data == 'l1':
            y_l1 = y
        else:
            y_l1 = y / y_scale
        # l1 correlation
        # --------------
        # E|X| = 1 --> E√XX = 1
        yy = np.multiply(y_l1, y_pred_l1, out=y_pred_l1).sum(0)
        sign = np.sign(yy)
        np.abs(yy, out=yy)
        yy **= 0.5
        yy *= sign
        return yy.mean()


EVALUATORS = {
    'l1': L1,
    'l2': L2,
    'r': Correlation,
    'r_rank': RankCorrelation,
    'vec-l1': VectorL1,
    'vec-l2': VectorL2,
    'vec-corr': VectorCorrelation,
    'vec-corr-l1': VectorCorrelationL1,
}


def get_evaluators(
        keys: List[str],
        data: RevCorrData,
        segments: List[np.ndarray] = None,  # evaluations for different test-segments
) -> (List[Evaluator], List[Evaluator], List[Evaluator]):
    evaluators = [EVALUATORS[key](data, segments) for key in keys]
    # split into scalar and vector evaluators
    evaluators_s = [e for e in evaluators if not e.vector]
    evaluators_v = [e for e in evaluators if e.vector]
    return evaluators, evaluators_s, evaluators_v



# make mne backwards-compatibility
import packaging.version

import mne


MNE_VERSION = packaging.version.parse(mne.__version__)
V0_19 = packaging.version.parse('0.19')
V0_22 = packaging.version.parse('0.22')

# 0.22 renamed read_ch_connectivity → read_ch_adjacency
if MNE_VERSION < V0_22:
    mne.channels.find_ch_adjacency = mne.channels.find_ch_connectivity
    mne.channels.read_ch_adjacency = mne.channels.read_ch_connectivity
    mne.spatial_src_adjacency = mne.spatial_src_connectivity



from colormath.color_objects import sRGBColor, HSVColor
from colormath.color_conversions import convert_color
from matplotlib.colors import to_rgb


def adjust_hsv(color, h=0., s=0., v=0.):
    hsv = convert_color(sRGBColor(*to_rgb(color)), HSVColor)
    hsv.hsv_h += h
    hsv.hsv_s += s
    hsv.hsv_v += v
    return convert_color(hsv, sRGBColor).get_value_tuple()


def set_hsv(color, h=None, s=None, v=None):
    hsv = convert_color(sRGBColor(*to_rgb(color)), HSVColor)
    if h is not None:
        hsv.hsv_h = h
    if s is not None:
        hsv.hsv_s = s
    if v is not None:
        hsv.hsv_v = v
    return convert_color(hsv, sRGBColor).get_value_tuple()



"""GUI to detect and load stc files and experimental conditions

Prompts user for experiment design information, and upon submission
loads stcs into an ``eelbrain.Dataset`` via a ``DatasetSTCLoader``
instance.
"""

import os
import wx

from . import get_app
from .frame import EelbrainFrame
from .._io.stc_dataset import DatasetSTCLoader


TEST_MODE = False


class STCLoaderFrame(EelbrainFrame):

    add_params = dict(proportion=0, flag=wx.EXPAND | wx.ALL, border=10)

    def __init__(self, parent):
        super().__init__(parent, wx.ID_ANY, "Find and Load STCs")
        self.loader = None
        self.factor_name_ctrls = []
        self.InitUI()
        self.Show()
        self.Center()
        self.Raise()

    def InitUI(self):
        self.SetMinSize((500, -1))
        self.status = self.CreateStatusBar(1)
        self.sizer = wx.BoxSizer(wx.VERTICAL) # top-level sizer
        data_title = TitleSizer(self, "MEG/MRI Information")
        self.sizer.Add(data_title, **self.add_params)
        dir_label = wx.StaticText(self, label=".stc Data Directory")
        self.dir_ctl = wx.DirPickerCtrl(self)
        if self.dir_ctl.HasTextCtrl():
            self.dir_ctl.SetTextCtrlProportion(5)
            self.dir_ctl.SetPickerCtrlProportion(1)
        vsizer = wx.BoxSizer(wx.VERTICAL)
        vsizer.Add(dir_label, 0, wx.BOTTOM, 2)
        vsizer.Add(self.dir_ctl, 0, wx.EXPAND)
        self.sizer.Add(vsizer, **self.add_params)
        self.mri_panel = MRIPanel(self)
        self.sizer.Add(self.mri_panel, **self.add_params)
        self.design_title = TitleSizer(self, "Experiment Design")
        self.sizer.Add(self.design_title, **self.add_params)
        self.design_title.ctl.Hide()
        self.factor_sizer = wx.BoxSizer(wx.HORIZONTAL)
        self.sizer.Add(self.factor_sizer, **self.add_params)
        bottom_sizer = wx.BoxSizer(wx.HORIZONTAL)
        self.attach_ds = wx.Button(self, wx.ID_ANY, "Return Dataset")
        self.launch_stats = wx.Button(self, wx.ID_ANY, "Launch Stats GUI")
        self.attach_ds.Disable()
        self.launch_stats.Disable()
        bottom_sizer.Add(self.attach_ds, 0, wx.ALIGN_RIGHT)
        bottom_sizer.Add(self.launch_stats, 0, wx.ALIGN_RIGHT)
        self.sizer.Add(bottom_sizer, 0, wx.ALIGN_RIGHT | wx.ALL, 10)
        self.sizer.Layout()
        self.SetSizer(self.sizer)
        self.sizer.Fit(self)

        self.Bind(wx.EVT_DIRPICKER_CHANGED, self.OnDirChange, self.dir_ctl)
        self.Bind(wx.EVT_BUTTON, self.OnAttachDataset, self.attach_ds)
        self.Bind(wx.EVT_BUTTON, self.OnLaunchStats, self.launch_stats)

    def OnDirChange(self, dir_picker_evt):
        """Create dataset loader and display level/factor names"""
        self.factor_sizer.Clear(True)
        path = dir_picker_evt.GetPath()
        try:
            self.loader = DatasetSTCLoader(path)
            self.status.SetStatusText("")
        except ValueError as err:
            self.status.SetStatusText(str(err))
            return
        self.DisplayLevels(self.loader.levels)
        self.attach_ds.Enable()
        self.launch_stats.Enable()

    def DisplayLevels(self, levels):
        """Show level names and factor name input for each factor"""
        self.design_title.ctl.Show()
        self.factor_name_ctrls = []
        for i, lvls in enumerate(levels):
            panel = FactorPanel(self, lvls, i)
            # self.factor_sizer.Add(panel, 0, wx.EXPAND | wx.RIGHT, 30)
            self.factor_sizer.Add(panel, 0, wx.EXPAND | wx.RIGHT, 30)
            self.factor_name_ctrls.append(panel.factor_ctl)
        self.factor_sizer.Layout()
        self.sizer.Layout()
        self.sizer.Fit(self)

    def _get_factor_names(self):
        return [c.GetValue() for c in self.factor_name_ctrls]

    def _get_stc_kwargs(self):
        kw = dict()
        kw["subjects_dir"] = self.mri_panel.mri_dir.GetPath()
        kw["subject"] = self.mri_panel.mri_subj.GetValue()
        kw["src"] = self.mri_panel.mri_src.GetValue()
        return kw

    def _get_dataset(self):
        names = self._get_factor_names()
        self.loader.set_factor_names(names)
        stc_kw = self._get_stc_kwargs()
        return self.loader.make_dataset(**stc_kw)

    def OnAttachDataset(self, evt):
        ds = self._get_dataset()
        get_app().Attach(ds, "Dataset from STCs", "ds", self)
        self.Close()

    def OnLaunchStats(self, evt):
        raise NotImplementedError()


class FactorPanel(wx.Panel):
    """Panel to display a factor and its level names"""
    def __init__(self, parent, levels, idx, editable=True):
        super(FactorPanel, self).__init__(parent)
        if editable:
            self.factor_ctl = wx.TextCtrl(self, value="factor_%d" % idx)
        else:
            self.factor_ctl = wx.StaticText(self, label="factor_%d" % idx)
            self.factor_ctl.SetFont(wx.Font.Bold(self.factor_ctl.GetFont()))
        level_names = ["- " + i for i in levels]
        level_ctl = wx.StaticText(self)
        level_ctl.SetLabel("\n".join(level_names))
        sizer = wx.BoxSizer(wx.VERTICAL)
        sizer.Add(self.factor_ctl)
        sizer.Add(level_ctl, 0, wx.EXPAND | wx.TOP, 10)
        sizer.Layout()
        self.SetSizer(sizer)


class MRIPanel(wx.Panel):
    """Panel containing MRI input fields (dir, subject, src)"""
    def __init__(self, parent):
        super(MRIPanel, self).__init__(parent)
        sizer = wx.StaticBoxSizer(wx.VERTICAL, self, "")
        # check for subjects dir in environment
        sdir = os.environ.get("SUBJECTS_DIR")
        if sdir is None:
            sdir = "/Applications/freesurfer/subjects"
        # directory control for MRI subjects dir, default freesurfer
        dir_ctl = wx.DirPickerCtrl(self, path=sdir)
        if dir_ctl.HasTextCtrl():
            dir_ctl.SetTextCtrlProportion(5)
            dir_ctl.SetPickerCtrlProportion(1)
        # text control for MRI subject, default 'fsaverage'
        subj_ctl = wx.TextCtrl(self, value="fsaverage")
        # dropdown to choose source space sampling, default ico-4
        srcs = ["ico-%d" % i for i in range(2, 7)] + ["oct-%d" % i for i in range(2, 7)] + ["all"]
        src_ctl = wx.ComboBox(self, choices=srcs, value="ico-4")
        # attach controls to panel for use in loader
        self.mri_dir = dir_ctl
        self.mri_subj = subj_ctl
        self.mri_src = src_ctl
        sizer.Add(wx.StaticText(self, label="MRI Directory"), 0, wx.BOTTOM, 2)
        sizer.Add(dir_ctl, 0, wx.EXPAND | wx.BOTTOM, 5)
        hs = wx.BoxSizer(wx.HORIZONTAL)
        for label, ctl in zip(("Subject", "Source Space"),
                              (subj_ctl, src_ctl)):
            vs = wx.BoxSizer(wx.VERTICAL)
            vs.Add(wx.StaticText(self, label=label), 0)
            vs.Add(ctl, 0)
            hs.Add(vs, 0, wx.RIGHT, 15)
        sizer.Add(hs, 0, wx.BOTTOM, 5)
        sizer.Layout()
        self.SetSizer(sizer)


class TitleSizer(wx.BoxSizer):
    def __init__(self, parent, title):
        super().__init__(wx.HORIZONTAL)
        self.ctl = wx.StaticText(parent, label=title)
        font = wx.Font(14, wx.DEFAULT, wx.NORMAL, wx.BOLD)
        self.ctl.SetFont(font)
        self.Add(self.ctl)



"""Statistical tests for multidimensional data in :class:`NDVar` objects"""
__test__ = False
from ._utils import _deprecated_alias

from ._stats.testnd import NDTest, MultiEffectNDTest, Correlation, TTestOneSample, TTestIndependent, TTestRelated, TContrastRelated, ANOVA, Vector, VectorDifferenceRelated
from ._stats.spm import LM, LMGroup

# for backwards compatibility
corr = _deprecated_alias('corr', Correlation, '0.34')
ttest_1samp = _deprecated_alias('ttest_1samp', TTestOneSample, '0.34')
ttest_ind = _deprecated_alias('ttest_ind', TTestIndependent, '0.34')
ttest_rel = _deprecated_alias('ttest_rel', TTestRelated, '0.34')
t_contrast_rel = _deprecated_alias('t_contrast_rel', TTestIndependent, '0.34')
anova = _deprecated_alias('anova', ANOVA, '0.34')



# Author: Christian Brodbeck <christianbrodbeck@nyu.edu>
"""Configure Eelbrain"""
import multiprocessing
import os
import sys

from matplotlib.colors import to_rgb

from ._utils import IS_OSX


SUPPRESS_WARNINGS = True
CONFIG = {
    'n_workers': multiprocessing.cpu_count(),
    'eelbrain': True,
    'autorun': None,
    'show': True,
    'format': 'svg',
    'figure_background': 'white',
    'prompt_toolkit': True,
    'animate': True,
    'nice': 0,
    'tqdm': False,  # disable=CONFIG['tqdm']
}

# Python 3.8 switched default to spawn, which makes pytest hang  (https://docs.python.org/3/whatsnew/3.8.html#multiprocessing)
method = multiprocessing.get_start_method()
if sys.version_info.minor >= 8 and IS_OSX:
    method = 'fork'
mpc = multiprocessing.get_context(method)


def configure(
        n_workers=None,
        frame=None,
        autorun=None,
        show=None,
        format=None,
        figure_background=None,
        prompt_toolkit=None,
        animate=None,
        nice=None,
        tqdm=None,
):
    """Set basic configuration parameters for the current session

    Parameters
    ----------
    n_workers : bool | int
        Number of worker processes to use in multiprocessing enabled
        computations. ``False`` to disable multiprocessing. ``True`` (default)
        to use as many processes as cores are available. Negative numbers to use
        all but n available CPUs.
    frame : bool
        Open figures in the Eelbrain application. This provides additional
        functionality such as copying a figure to the clipboard. If False, open
        figures as normal matplotlib figures.
    autorun : bool
        When a figure is created, automatically enter the GUI mainloop. By
        default, this is True when the figure is created in interactive mode
        but False when the figure is created in a script (in order to run the
        GUI at a specific point in a script, call :func:`eelbrain.gui.run`).
    show : bool
        Show plots on the screen when they're created (disable this to create
        plots and save them without showing them on the screen).
    format : str
        Default format for plots (for example "png", "svg", ...).
    figure_background : bool | matplotlib color
        While :mod:`matplotlib` uses a gray figure background by default,
        Eelbrain uses white. Set this parameter to ``False`` to use the default
        from :attr:`matplotlib.rcParams`, or set it to a valid matplotblib
        color value to use an arbitrary color. ``True`` to revert to the default
        white.
    prompt_toolkit : bool
        In IPython 5, prompt_toolkit allows running the GUI main loop in
        parallel to the Terminal, meaning that the IPython terminal and GUI
        windows can be used without explicitly switching between Terminal and
        GUI. This feature is enabled by default, but can be disabled by setting
        ``prompt_toolkit=False``.
    animate : bool
        Animate plot navigation (default True).
    nice : int [-20, 19]
        Scheduling priority for muliprocessing (larger number yields more to
        other processes; negative numbers require root privileges).
    tqdm : bool
        Enable or disable :mod:`tqdm` progress bars.
    """
    # don't change values before raising an error
    new = {}
    if n_workers is not None:
        if n_workers is True:
            new['n_workers'] = multiprocessing.cpu_count()
        elif n_workers is False:
            new['n_workers'] = 0
        elif isinstance(n_workers, int):
            cpu_count = multiprocessing.cpu_count()
            if n_workers < 0:
                if cpu_count - n_workers < 1:
                    raise ValueError(f"n_workers={n_workers}, but only {cpu_count} CPUs are available")
                new['n_workers'] = cpu_count - n_workers
            else:
                new['n_workers'] = n_workers
        else:
            raise TypeError("n_workers=%r" % (n_workers,))
    if frame is not None:
        new['eelbrain'] = bool(frame)
    if autorun is not None:
        new['autorun'] = bool(autorun)
    if show is not None:
        new['show'] = bool(show)
    if format is not None:
        new['format'] = format.lower()
    if figure_background is not None:
        if figure_background is True:
            figure_background = 'white'
        elif figure_background is not False:
            to_rgb(figure_background)
        new['figure_background'] = figure_background
    if prompt_toolkit is not None:
        new['prompt_toolkit'] = bool(prompt_toolkit)
    if animate is not None:
        new['animate'] = bool(animate)
    if nice is not None:
        nice = int(nice)
        if not -20 <= nice < 20:
            raise ValueError("nice=%i; needs to be in range [-20, 19]" % (nice,))
        elif nice < 0 and not os.getuid() == 0:
            raise ValueError("nice=%i; values < 0 require root privileges" % (nice,))
        new['nice'] = nice
    if tqdm is not None:
        new['tqdm'] = not tqdm

    CONFIG.update(new)



# Author: Christian Brodbeck <christianbrodbeck@nyu.edu>
from math import floor

import numpy as np
from numpy import newaxis


FULL_SLICE = slice(None)
FULL_AXIS_SLICE = (FULL_SLICE,)
INT_TYPES = (int, np.integer)


def digitize_index(index: float, values: np.ndarray, tol: float=None):
    """Locate a scalar ``index`` on ``values``

    Parameters
    ----------
    index : scalar
        Index to locate on ``values``.
    values : array
        1-dimensional array on which to locate ``index``.
    tol : float
        Tolerance for suppressing an IndexError when index only falls near a
        value.
    """
    i = int(np.digitize(index, values, True))
    if index == values[i]:
        return i
    elif not tol:
        raise IndexError(f"{index}: does not match any value")
    elif i > 0 and values[i] - index > index - values[i - 1]:
        i -= 1
    diff = abs(index - values[i])
    if i == 0:
        if len(values) == 1:
            if diff <= abs(values[0]) * tol:
                return i
            else:
                raise IndexError("Index %r outside of tolerance from only "
                                 "value %r" % (index, values[0]))
        elif diff <= abs(values[1] - values[0]) * tol:
            return i
        else:
            raise IndexError("Index %r outside of tolerance" % (index,))
    elif diff <= abs(values[i] - values[i - 1]) * tol:
        return i
    else:
        raise IndexError("Index %r outside of tolerance" % (index,))


def digitize_slice_endpoint(index, values):
    """Locate a scalar slice endpoint on ``values``

    (Whenever index is between two values, move to the larger)

    Parameters
    ----------
    index : scalar
        Index to locate on ``values``.
    values : array
        1-dimensional array on which to locate ``index``.
    """
    return int(np.digitize(index, values, True))


def index(index, at: int):
    return FULL_AXIS_SLICE * at + (index,)


def index_to_int_array(index, n):
    if isinstance(index, np.ndarray):
        if index.dtype.kind == 'i':
            return index
        elif index.dtype.kind == 'b':
            return np.flatnonzero(index)
    return np.arange(n)[index]


def index_length(index, n):
    "Length of an array index (number of selected elements)"
    if isinstance(index, slice):
        start = index.start or 0
        if start < 0:
            start = n - start
        stop = n if index.stop is None else index.stop
        if stop < 0:
            stop = n - stop
        step = index.step or 1
        return floor((stop - start) / step)

    if not isinstance(index, np.ndarray):
        index = np.asarray(index)
    if index.dtype.kind == 'b':
        if len(index) > n:
            index = index[:n]
        return index.sum()
    elif index.dtype.kind in 'iu':
        return len(index)
    else:
        raise TypeError("index %r" % (index,))


def apply_numpy_index(data, index):
    "Apply numpy index to non-numpy sequence"
    if isinstance(index, (int, slice)):
        return data[index]
    elif isinstance(index, list):
        return (data[i] for i in index)
    array = np.asarray(index)
    assert array.ndim == 1, "Index must be 1 dimensional, got %r" % (index,)
    if array.dtype.kind == 'i':
        return (data[i] for i in array)
    elif array.dtype.kind == 'b':
        assert len(array) == len(data), "Index must have same length as data"
        return (d for d, i in zip(data, array) if i)
    raise TypeError("Invalid numpy-like index: %r" % (index,))


def slice_to_arange(s, length):
    """Convert a slice into a numerical index

    Parameters
    ----------
    s : slice
        Slice.
    length : int
        Length of the target for the index (only needed if ``slice.stop`` is
        None).

    Returns
    -------
    arange : array of int
        Numerical index equivalent to slice.
    """
    if s.start is None:
        start = 0
    else:
        start = s.start

    if s.stop is None:
        stop = length
    else:
        stop = s.stop

    return np.arange(start, stop, s.step)


def aindex(axis, index):
    return FULL_AXIS_SLICE * axis + (index,)


def aslice(axis, start=None, stop=None, step=None):
    return FULL_AXIS_SLICE * axis + (slice(start, stop, step),)


def take_slice(x, axis, start=None, stop=None, step=None):
    """Like :func:`numpy.take` but with a slice"""
    return x[aslice(axis, start, stop, step)]


def deep_array(data, axis, ndim):
    "Array in which ``data`` lies along ``axis`` of ``ndim`` dimensions"
    if axis < 0:
        axis += ndim
        assert axis >= 0
    out = np.array(data)
    out.shape = (1,) * axis + out.shape + (1,) * (ndim - axis - out.ndim)
    return out



# Author: Christian Brodbeck <christianbrodbeck@nyu.edu>
"""I/O for MNE"""
from collections.abc import Iterable
import fnmatch
from itertools import zip_longest
from logging import getLogger
from math import floor
import os
from pathlib import Path
import re
from typing import Any, List, Optional, Sequence, Tuple, Union
import warnings

import numpy as np

import mne
from mne.source_estimate import _BaseSourceEstimate
from mne.io.constants import FIFF
from mne.io.kit.constants import KIT

from .. import _info
from .._types import PathArg
from .._data_obj import Factor, Var, NDVar, Dataset, Case, Sensor, Space, SourceSpace, VolumeSourceSpace, UTS, _matrix_graph
from .._info import BAD_CHANNELS
from .._text import n_of
from .._utils import ui, as_list
from ..mne_fixes import MNE_EVOKED, MNE_RAW, MNE_VOLUME_STC


try:  # mne >= 0.19
    from mne.io.kit.constants import KIT_NEIGHBORS
except ImportError:
    KIT_NEIGHBORS = {
        KIT.SYSTEM_NYU_2008: 'KIT-157',
        KIT.SYSTEM_NYU_2009: 'KIT-157',
        KIT.SYSTEM_NYU_2010: 'KIT-157',
        KIT.SYSTEM_NYUAD_2011: 'KIT-208',
        KIT.SYSTEM_NYUAD_2012: 'KIT-208',
        KIT.SYSTEM_NYUAD_2014: 'KIT-208',
        KIT.SYSTEM_UMD_2004: 'KIT-UMD-1',
        KIT.SYSTEM_UMD_2014_07: 'KIT-UMD-2',
        KIT.SYSTEM_UMD_2014_12: 'KIT-UMD-3',
    }

BaselineArg = Optional[Tuple[Optional[float], Optional[float]]]
PicksArg = Any


def mne_raw(path=None, proj=False, **kwargs):
    """Load a :class:`mne.io.Raw` object

    Parameters
    ----------
    path : None | str
        path to a raw fiff or sqd file. If no path is supplied, a file can be
        chosen from a file dialog.
    proj : bool | str
        Add projections from a separate file to the Raw object.
        **``False``**: No proj file will be added.
        **``True``**: ``'{raw}*proj.fif'`` will be used.
        ``'{raw}'`` will be replaced with the raw file's path minus '_raw.fif',
        and '*' will be expanded using fnmatch. If multiple files match the
        pattern, a ValueError will be raised.
        **``str``**: A custom path template can be provided, ``'{raw}'`` and
        ``'*'`` will be treated as with ``True``.
    kwargs
        Additional keyword arguments are forwarded to :class:`mne.io.Raw`
        initialization.
    """
    if path is None:
        path = ui.ask_file("Pick a raw data file", "Pick a raw data file",
                           [('Functional image file (*.fif)', '*.fif'),
                            ('KIT Raw File (*.sqd,*.con', '*.sqd;*.con')])
        if not path:
            return

    if isinstance(path, str):
        path = Path(path)
    if isinstance(path, Path):
        ext = path.suffix.lower()
        if ext.startswith('.fif'):
            raw = mne.io.read_raw_fif(path, **kwargs)
        elif ext in ('.sqd', '.con'):
            raw = mne.io.read_raw_kit(path, **kwargs)
        else:
            raise ValueError(f"{path}: Unknown extension")
    elif isinstance(path, Iterable):
        # MNE Raw supports list of file-names
        raw = mne.io.read_raw_fif(path, **kwargs)
    else:
        raise TypeError("path=%r" % (path,))

    if proj:
        if proj is True:
            proj = '{raw}*proj.fif'

        if '{raw}' in proj:
            raw_file = raw.filenames[0]
            raw_root, _ = os.path.splitext(raw_file)
            raw_root = raw_root.rstrip('raw')
            proj = proj.format(raw=raw_root)

        if '*' in proj:
            head, tail = os.path.split(proj)
            names = fnmatch.filter(os.listdir(head), tail)
            if len(names) == 1:
                proj = os.path.join(head, names[0])
            else:
                if len(names) == 0:
                    err = "No file matching %r"
                else:
                    err = "Multiple files matching %r"
                raise ValueError(err % proj)

        # add the projections to the raw file
        proj = mne.read_proj(proj)
        raw.add_proj(proj, remove_existing=True)

    return raw


def events(raw=None, merge=None, proj=False, name=None, bads=None,
           stim_channel=None, events=None, annotations=None, **kwargs):
    """
    Load events from a raw fiff file.

    Parameters
    ----------
    raw : str(path) | None | mne Raw
        The raw fiff file from which to extract events (if raw and events are
        both ``None``, a file dialog will be displayed to select a raw file).
    merge : int
        Merge stimulus channel steps occurring in neighboring samples.
        The integer value indicates across how many samples events should
        be merged, and the sign indicates in which direction they should be
        merged (negative means towards the earlier event, positive towards
        the later event). By default, this parameter is based on the data:
        -1 for KIT data, 0 otherwise.
        The main reason for merging events is an artifact from analog event
        recording systems. If events are recorded in an analog channel,
        event onsets can be blurred (what should be ``0, 0, 0, 1, 1, 1, ...``,
        can look like ``0, 0, 0, 0.5, 1, 1, ...``; ``merge=-1`` would
        turn the latter into the former).
    proj : bool | str
        Path to the projections file that will be loaded with the raw file.
        ``'{raw}'`` will be expanded to the raw file's path minus extension.
        With ``proj=True``, ``'{raw}_*proj.fif'`` will be used,
        looking for any projection file starting with the raw file's name.
        If multiple files match the pattern, a ValueError will be raised.
    name : str | None
        A name for the Dataset. If ``None``, the raw filename will be used.
    bads : None | list
        Specify additional bad channels in the raw data file (these are added
        to the ones that are already defined in the raw file).
    stim_channel : None | string | list of string
        Name of the stim channel or all the stim channels
        affected by the trigger. If None, the config variables
        'MNE_STIM_CHANNEL', 'MNE_STIM_CHANNEL_1', 'MNE_STIM_CHANNEL_2',
        etc. are read. If these are not found, it will default to
        'STI 014'.
    events : None | str
        If events are stored in a fiff file separate from the Raw object, the
        path to the events file can be supplied here. The events in the Dataset
        will reflect the event sin the events file rather than the raw file.
    annotations : bool
        Generate events from annotations instead of the stim channel (by
        default, annotations are used when present).
    **
        Keyword arguments for loading the raw file (see
        :func:`mne.io.read_raw_kit` or :func:`mne.io.read_raw_kit`).

    Returns
    -------
    events : Dataset
        A Dataset with the following variables:
         - *i_start*: the index of the event in the raw file.
         - *trigger*: the event value/id.
         - *event*: the event label (only for events from annotations).
        The Dataset's info dictionary contains the following values:
         - *raw*: the mne Raw object.

    """
    if (raw is None and events is None) or isinstance(raw, (str, Path)):
        raw = mne_raw(raw, proj=proj, **kwargs)

    if bads is not None and raw is not None :
        raw.info['bads'].extend(bads)

    if name is None and raw is not None:
        raw_path = raw.filenames[0]
        if isinstance(raw_path, str):
            name = os.path.basename(raw_path)
        else:
            name = None

    labels = None
    if events is None:
        if annotations is None:
            regex = re.compile('(bad|edge)', re.IGNORECASE)
            index = [not regex.match(desc) for desc in raw.annotations.description]
            if any(index):
                annotations = raw.annotations[index]

        if annotations:
            evts, event_ids = mne.events_from_annotations(raw)
        else:
            raw.load_data()
            if merge is None:
                if raw.info.get('kit_system_id') is None:
                    merge = 0
                else:
                    merge = -1
            evts = mne.find_stim_steps(raw, merge=merge, stim_channel=stim_channel)
            evts = evts[np.flatnonzero(evts[:, 2])]
            event_ids = getattr(raw, 'event_id', None)

        if event_ids:
            labels = {event_id: label for label, event_id in event_ids.items()}
    else:
        evts = mne.read_events(events)

    ds = Dataset({
        'i_start': Var(evts[:, 0]),
        'trigger': Var(evts[:, 2]),
    }, name, info={'raw': raw})
    if labels is not None:
        ds['event'] = Factor(ds['trigger'], labels=labels)
    return ds


def find_mne_channel_types(info):
    chs = {(ch['kind'], ch['unit']) for ch in info['chs']}
    types = {ch[0] for ch in chs}
    if FIFF.FIFFV_MEG_CH in types:
        units = {ch[1] for ch in chs if ch[0] == FIFF.FIFFV_MEG_CH}
        mag = FIFF.FIFF_UNIT_T in units
        grad = FIFF.FIFF_UNIT_T_M in units
    else:
        mag = grad = False
    eeg = FIFF.FIFFV_EEG_CH in types
    eog = FIFF.FIFFV_EOG_CH in types
    out = []
    if mag:
        out.append('mag')
    if grad:
        out.append('planar1')
        out.append('planar2')
    if eeg:
        out.append('eeg')
    if eog:
        out.append('eog')
    return out


def _guess_ndvar_data_type(info):
    """Guess which type of data to extract from an mne object.

    Checks for the presence of channels in that order: "mag", "eeg", "grad".
    If none are found, a ValueError is raised.

    Parameters
    ----------
    info : dict
        MNE info dictionary.

    Returns
    -------
    data : str
        Kind of data to extract
    """
    data_types = find_mne_channel_types(info)
    if data_types:
        return data_types[0]
    else:
        raise ValueError("No MEG, EEG or EOG channel found in data")


def _picks(info, data, exclude) -> np.ndarray:
    if data is None:
        data = _guess_ndvar_data_type(info)
    if data == 'eeg':
        meg = False
        eeg = True
        eog = False
    elif data == 'eog':
        meg = False
        eeg = False
        eog = True
    elif data == 'eeg&eog':
        meg = False
        eeg = True
        eog = True
    elif data in ['grad', 'mag', 'planar1', 'planar2']:
        meg = data
        eeg = False
        eog = False
    elif data is True:
        meg = True
        eeg = True
        eog = False
    else:
        raise ValueError(f"data={data!r} (needs to be one of: eeg, eog, mag, grad, planar1, planar2)")
    return mne.pick_types(info, meg, eeg, False, eog, ref_meg=False, exclude=exclude)


def _ndvar_epochs_reject(data, reject):
    if reject:
        if not np.isscalar(reject):
            err = ("Reject must be scalar (rejection threshold); got %s." %
                   repr(reject))
            raise ValueError(err)
        reject = {data: reject}
    else:
        reject = None
    return reject


def _sensor_info(data, vmax, mne_info, user_info=None, mult=1):
    if data == 'eeg' or data == 'eog' or data == 'eeg&eog':
        info = _info.for_eeg(vmax, mult)
        summary_vmax = 0.1 * vmax if vmax else None
        summary_info = _info.for_eeg(summary_vmax, mult)
    elif data == 'mag':
        info = _info.for_meg(vmax, mult)
        summary_vmax = 0.1 * vmax if vmax else None
        summary_info = _info.for_meg(summary_vmax, mult)
    elif data in ['grad', 'planar1', 'planar2']:
        info = _info.for_meg(vmax, mult, 'T/cm', '∆U')
        summary_vmax = 0.1 * vmax if vmax else None
        summary_info = _info.for_meg(summary_vmax, mult, 'T/cm', '∆U')
    else:
        raise ValueError("data=%r" % data)
    info.update(proj='z root', samplingrate=mne_info['sfreq'], summary_info=summary_info)
    if user_info:
        info.update(user_info)
    return info


def epochs(ds, tmin=-0.1, tmax=None, baseline=None, decim=1, mult=1, proj=False,
           data=None, reject=None, exclude='bads', info=None, name=None,
           raw=None, sensors=None, i_start='i_start', tstop=None, sysname=None):
    """
    Load epochs as :class:`NDVar`.

    Parameters
    ----------
    ds : Dataset
        Dataset containing a variable which defines epoch cues (i_start).
    tmin : scalar
        First sample to include in the epochs in seconds (Default is -0.1).
    tmax : scalar
        Last sample to include in the epochs in seconds (Default 0.6; use
        ``tstop`` instead to specify index exclusive of last sample).
    baseline : (float, float) | None
        Time interval for baseline correction. ``(tmin, tmax)`` tuple in
        seconds, or ``None`` to use all the data (e.g., ``(None, 0)`` uses all
        the data from the beginning of the epoch up to ``t = 0``). Set to
        ``None`` for no baseline correction (default).
    decim : int
        Downsample the data by this factor when importing. ``1`` means no
        downsampling. Note that this function does not low-pass filter
        the data. The data is downsampled by picking out every
        n-th sample (see `Wikipedia <http://en.wikipedia.org/wiki/Downsampling>`_).
    mult : scalar
        multiply all data by a constant.
    proj : bool
        mne.Epochs kwarg (subtract projections when loading data)
    data : 'eeg' | 'mag' | 'grad'
        Which data channels data to include (default based on channels in data).
    reject : None | scalar
        Threshold for rejecting epochs (peak to peak). Requires a for of
        mne-python which implements the Epochs.model['index'] variable.
    exclude : list of string | str
        Channels to exclude (:func:`mne.pick_types` kwarg).
        If 'bads' (default), exclude channels in info['bads'].
        If empty do not exclude any.
    info : None | dict
        Entries for the ndvar's info dict.
    name : str
        name for the new NDVar.
    raw : None | mne Raw
        Raw file providing the data; if ``None``, ``ds.info['raw']`` is used.
    sensors : None | Sensor
        The default (``None``) reads the sensor locations from the fiff file.
        If the fiff file contains incorrect sensor locations, a different
        Sensor instance can be supplied through this kwarg.
    i_start : str
        name of the variable containing the index of the events.
    tstop : scalar
        Alternative to ``tmax``: While ``tmax`` specifies the last samples to 
        include, ``tstop`` can be used to specify the epoch time excluding the 
        last time point (i.e., standard Python/Eelbrain indexing convention).
        For example, at 100 Hz the epoch with ``tmin=-0.1, tmax=0.4`` will have 
        51 samples, while the epoch specified with ``tmin=-0.1, tstop=0.4`` will
        have 50 samples.
    sysname : str
        Name of the sensor system to load sensor connectivity (e.g. 'neuromag',
        inferred automatically for KIT data converted with a recent version of
        MNE-Python).

    Returns
    -------
    epochs : NDVar
        The epochs as NDVar object.
    """
    if raw is None:
        raw = ds.info['raw']

    if data is None:
        data = _guess_ndvar_data_type(raw.info)
    picks = _picks(raw.info, data, exclude)
    reject = _ndvar_epochs_reject(data, reject)

    epochs_ = mne_epochs(ds, tmin, tmax, baseline, i_start, raw, decim=decim,
                         picks=picks, reject=reject, proj=proj, tstop=tstop)
    ndvar = epochs_ndvar(epochs_, name, data, mult=mult, info=info,
                         sensors=sensors, sysname=sysname)

    if len(epochs_) == 0:
        raise RuntimeError(f"No events left in {raw.filenames[0]}")
    return ndvar


def add_epochs(ds, tmin=-0.1, tmax=0.6, baseline=None, decim=1, mult=1,
               proj=False, data=None, reject=None, exclude='bads', info=None,
               name="meg", raw=None, sensors=None, i_start='i_start',
               sysname=None, tstop=None):
    """
    Load epochs and add them to a dataset as :class:`NDVar`.

    Unless the ``reject`` argument is specified, ``ds``
    is modified in place. With ``reject``, a subset of ``ds`` is returned
    containing only those events for which data was loaded.

    Parameters
    ----------
    ds : Dataset
        Dataset containing a variable which defines epoch cues (i_start) and to
        which the epochs are added.
    tmin : scalar
        First sample to include in the epochs in seconds (Default is -0.1).
    tmax : scalar
        Last sample to include in the epochs in seconds (Default 0.6; use
        ``tstop`` instead to specify index exclusive of last sample).
    baseline : (float, float) | None
        Time interval for baseline correction. ``(tmin, tmax)`` tuple in
        seconds, or ``None`` to use all the data (e.g., ``(None, 0)`` uses all
        the data from the beginning of the epoch up to ``t = 0``). Set to
        ``None`` for no baseline correction (default).
    decim : int
        Downsample the data by this factor when importing. ``1`` means no
        downsampling. Note that this function does not low-pass filter
        the data. The data is downsampled by picking out every
        n-th sample (see `Wikipedia <http://en.wikipedia.org/wiki/Downsampling>`_).
    mult : scalar
        multiply all data by a constant.
    proj : bool
        mne.Epochs kwarg (subtract projections when loading data)
    data : 'eeg' | 'mag' | 'grad'
        Which data channels data to include (default based on channels in data).
    reject : None | scalar
        Threshold for rejecting epochs (peak to peak). Requires a for of
        mne-python which implements the Epochs.model['index'] variable.
    exclude : list of string | str
        Channels to exclude (:func:`mne.pick_types` kwarg).
        If 'bads' (default), exclude channels in info['bads'].
        If empty do not exclude any.
    info : None | dict
        Entries for the ndvar's info dict.
    name : str
        name for the new NDVar.
    raw : None | mne Raw
        Raw file providing the data; if ``None``, ``ds.info['raw']`` is used.
    sensors : None | Sensor
        The default (``None``) reads the sensor locations from the fiff file.
        If the fiff file contains incorrect sensor locations, a different
        Sensor instance can be supplied through this kwarg.
    i_start : str
        name of the variable containing the index of the events.
    sysname : str
        Name of the sensor system to load sensor connectivity (e.g. 'neuromag',
        inferred automatically for KIT data converted with a recent version of
        MNE-Python).
    tstop : scalar
        Alternative to ``tmax``: While ``tmax`` specifies the last samples to 
        include, ``tstop`` can be used to specify the epoch time excluding the 
        last time point (i.e., standard Python/Eelbrain indexing convention).
        For example, at 100 Hz the epoch with ``tmin=-0.1, tmax=0.4`` will have 
        51 samples, while the epoch specified with ``tmin=-0.1, tstop=0.4`` will
        have 50 samples.

    Returns
    -------
    ds : Dataset
        Dataset containing the epochs. If no events are rejected, ``ds`` is the
        same object as the input ``ds`` argument, otherwise a copy of it.
    """
    if raw is None:
        raw = ds.info['raw']

    if data is None:
        data = _guess_ndvar_data_type(raw.info)
    picks = _picks(raw.info, data, exclude)
    reject = _ndvar_epochs_reject(data, reject)

    epochs_ = mne_epochs(ds, tmin, tmax, baseline, i_start, raw, decim=decim,
                         picks=picks, reject=reject, proj=proj, tstop=tstop)
    ds = _trim_ds(ds, epochs_)
    ds[name] = epochs_ndvar(epochs_, name, data, mult=mult, info=info,
                            sensors=sensors, sysname=sysname)
    return ds


def add_mne_epochs(ds, tmin=-0.1, tmax=None, baseline=None, target='epochs', **kwargs):
    """
    Load epochs and add them to a dataset as :class:`mne.Epochs`.

    If, after loading, the Epochs contain fewer cases than the Dataset, a copy
    of the Dataset is made containing only those events also contained in the
    Epochs. Note that the Epochs are always loaded with ``preload==True``.

    If the Dataset's info dictionary contains a 'bad_channels' entry, those bad
    channels are added to the epochs.


    Parameters
    ----------
    ds : Dataset
        Dataset with events from a raw fiff file (i.e., created by
        load.fiff.events).
    tmin : scalar
        First sample to include in the epochs in seconds (Default is -0.1).
    tmax : scalar
        Last sample to include in the epochs in seconds (Default 0.6; use
        ``tstop`` instead to specify index exclusive of last sample).
    baseline : (float, float) | None
        Time interval for baseline correction. ``(tmin, tmax)`` tuple in
        seconds, or ``None`` to use all the data (e.g., ``(None, 0)`` uses all
        the data from the beginning of the epoch up to ``t = 0``). Set to
        ``None`` for no baseline correction (default).
    target : str
        Name for the Epochs object in the Dataset.
    ...
        See :func:`~eelbrain.load.fiff.mne_epochs`.
    """
    epochs_ = mne_epochs(ds, tmin, tmax, baseline, **kwargs)
    ds = _trim_ds(ds, epochs_)
    ds[target] = epochs_
    return ds


def _mne_events(ds=None, i_start='i_start', trigger='trigger'):
    """Convert events from a Dataset into mne events"""
    if isinstance(i_start, str):
        i_start = ds[i_start]

    N = len(i_start)

    if isinstance(trigger, str):
        trigger = ds[trigger]
    elif trigger is None:
        trigger = np.ones(N)

    events = np.empty((N, 3), dtype=np.int32)
    events[:, 0] = i_start.x
    events[:, 1] = 0
    events[:, 2] = trigger
    return events


def mne_epochs(ds, tmin=-0.1, tmax=None, baseline=None, i_start='i_start',
               raw=None, drop_bad_chs=True, picks=None, reject=None, tstop=None,
               decim=1, **kwargs):
    """Load epochs as :class:`mne.Epochs`.

    Parameters
    ----------
    ds : Dataset
        Dataset containing a variable which defines epoch cues (i_start).
    tmin : scalar
        First sample to include in the epochs in seconds (Default is -0.1).
    tmax : scalar
        Last sample to include in the epochs in seconds (Default 0.6; use
        ``tstop`` instead to specify index exclusive of last sample).
    baseline : (float, float) | None
        Time interval for baseline correction. ``(tmin, tmax)`` tuple in
        seconds, or ``None`` to use all the data (e.g., ``(None, 0)`` uses all
        the data from the beginning of the epoch up to ``t = 0``). Set to
        ``None`` for no baseline correction (default).
    i_start : str
        name of the variable containing the index of the events.
    raw : None | mne Raw
        If None, ds.info['raw'] is used.
    drop_bad_chs : bool
        Drop all channels in raw.info['bads'] form the Epochs. This argument is
        ignored if the picks argument is specified.
    picks, reject
        :class:`mne.Epochs` parameters.
    tstop : scalar
        Alternative to ``tmax``: While ``tmax`` specifies the last samples to 
        include, ``tstop`` can be used to specify the epoch time excluding the 
        last time point (i.e., standard Python/Eelbrain indexing convention).
        For example, at 100 Hz the epoch with ``tmin=-0.1, tmax=0.4`` will have 
        51 samples, while the epoch specified with ``tmin=-0.1, tstop=0.4`` will
        have 50 samples.
    ...
        :class:`mne.Epochs` parameters.
    """
    if baseline is False:
        baseline = None
    if raw is None:
        raw = ds.info['raw']
    if tmax is None:
        if tstop is None:
            tmax = 0.6
        else:
            sfreq = raw.info['sfreq'] / decim
            start_index = int(round(tmin * sfreq))
            stop_index = int(round(tstop * sfreq))
            tmax = tmin + (stop_index - start_index - 1) / sfreq
    elif tstop is not None:
        raise TypeError(f"tmax and tstop can not both be specified at the same time, got tmax={tmax}, tstop={tstop}")

    if drop_bad_chs and picks is None and raw.info['bads']:
        picks = mne.pick_types(raw.info, meg=True, eeg=True, eog=True, ref_meg=False)

    events = _mne_events(ds=ds, i_start=i_start)
    with warnings.catch_warnings():
        warnings.filterwarnings('ignore', 'The events passed to the Epochs constructor', RuntimeWarning)
        epochs = mne.Epochs(raw, events, None, tmin, tmax, baseline, picks, preload=True, reject=reject, decim=decim, **kwargs)
    if reject is None and len(epochs) != len(events):
        getLogger('eelbrain').warning("%s: MNE generated only %i Epochs for %i events. The raw file might end before the end of the last epoch.", raw.filenames[0], len(epochs), len(events))

    #  add bad channels from ds
    if BAD_CHANNELS in ds.info:
        invalid = []
        for ch_name in ds.info[BAD_CHANNELS]:
            if ch_name in raw.info['bads']:
                pass
            elif ch_name not in epochs.ch_names:
                invalid.append(ch_name)
            elif ch_name not in epochs.info['bads']:
                epochs.info['bads'].append(ch_name)
        if invalid:
            raise ValueError(f"{n_of(len(invalid), 'invalid channel')} in ds.info[{BAD_CHANNELS!r}]: {', '.join(invalid)}")

    return epochs


def sensor_dim(
        info: mne.Info,
        picks: np.ndarray = None,
        sysname: str = None,
        connectivity: Union[str, Sequence] = None,
):
    """Create a :class:`Sensor` dimension from an :class:`mne.Info` object.

    Parameters
    ----------
    info
        Measurement info dictionary (or mne-python object that has a ``.info``
        attribute that contains measurement info).
    picks
        Channel picks (integer array, as used in mne-python).
        By default all MEG and EEG channels are included.
    sysname
        Name of the sensor system to load sensor connectivity (e.g. 'neuromag',
        inferred automatically for KIT data converted with a recent version of
        MNE-Python).
    connectivity : str | list of (str, str) | array of int, (n_edges, 2)
        Connectivity between elements. Can be specified as:

        - ``"none"`` for no connections
        - list of connections (e.g., ``[('OZ', 'O1'), ('OZ', 'O2'), ...]``)
        - :class:`numpy.ndarray` of int, shape (n_edges, 2), to specify
          connections in terms of indices. Each row should specify one
          connection [i, j] with i < j. If the array's dtype is uint32,
          property checks are disabled to improve efficiency.
        - ``"grid"`` to use adjacency in the sensor names

        If unspecified, it is inferred from ``sysname`` if possible.

    Returns
    -------
    sensor_dim : Sensor
        Sensor dimension object.
    """
    if not isinstance(info, mne.Info):
        info_ = getattr(info, 'info', info)
        if not isinstance(info_, mne.Info):
            raise TypeError("No mne.Info object: %r" % (info,))
        info = info_

    if picks is None:
        picks = mne.pick_types(info, meg=True, eeg=True, ref_meg=False, exclude=())
    else:
        picks = np.asarray(picks, int)

    chs = [info['chs'][i] for i in picks]
    ch_locs = [ch['loc'][:3] for ch in chs]
    ch_names = [ch['ch_name'] for ch in chs]

    # use KIT system ID if available
    sysname = KIT_NEIGHBORS.get(info.get('kit_system_id'), sysname)
    if sysname and sysname.startswith('neuromag'):
        ch_unit = {ch['unit'] for ch in chs}
        if len(ch_unit) > 1:
            raise RuntimeError(f"More than one channel kind for sysname={sysname!r}: {tuple(ch_unit)}")
        ch_unit = ch_unit.pop()
        if ch_unit == FIFF.FIFF_UNIT_T_M:
            sysname = 'neuromag306planar'
        elif ch_unit == FIFF.FIFF_UNIT_T:
            sysname = 'neuromag306mag'
        elif ch_unit == FIFF.FIFF_UNIT_V:
            sysname = 'neuromag306eeg'
            if connectivity is None:
                connectivity = 'auto'
        else:
            raise ValueError(f"Unknown channel unit for sysname={sysname!r}: {ch_unit!r}")

    if connectivity is None:
        connectivity = sysname  # or 'auto'

    if connectivity in ('grid', 'none'):
        pass
    elif isinstance(connectivity, str):
        if connectivity == 'auto':
            ch_type = _guess_ndvar_data_type({'chs': chs})
            c_matrix, names = mne.channels.find_ch_adjacency(info, ch_type)
        else:
            c_matrix, names = mne.channels.read_ch_adjacency(connectivity)
            if connectivity.startswith('neuromag'):
                vec_ids = {name[-1] for name in ch_names}
                if len(vec_ids) > 1:
                    raise NotImplementedError("Connectivity for Neuromag vector data")
                vec_id = vec_ids.pop()
                # fix channel names
                names = [f'MEG {n[3:-1]}{vec_id}' for n in names]

        # fix channel order
        if names != ch_names:
            index = np.array([names.index(name) for name in ch_names])
            c_matrix = c_matrix[index][:, index]

        connectivity = _matrix_graph(c_matrix)
    elif connectivity in (None, False):
        connectivity = 'none'

    return Sensor(ch_locs, ch_names, sysname, connectivity=connectivity)


def variable_length_mne_epochs(
        ds: Dataset,
        tmin: float,
        tmax: Sequence[float],
        baseline: BaselineArg = None,
        allow_truncation: bool = False,
        picks: PicksArg = None,
        **kwargs,
) -> List[mne.Epochs]:
    """Load mne Epochs where each epoch has a different length

    Parameters
    ----------
    ds : Dataset
        Dataset containing a variable which defines epoch cues (i_start).
    tmin : scalar
        First sample to include in the epochs in seconds (Default is -0.1).
    tmax : sequence of scalar
        Last sample to include in each epoch in seconds.
    baseline : (float, float) | None
        Time interval for baseline correction. ``(tmin, tmax)`` tuple in
        seconds, or ``None`` to use all the data (e.g., ``(None, 0)`` uses all
        the data from the beginning of the epoch up to ``t = 0``). Set to
        ``None`` for no baseline correction (default).
    allow_truncation : bool
        If a ``tmax`` value falls outside the data available in ``raw``,
        automatically truncate the epoch (by default this raises a
        ``ValueError``).
    ...
        :class:`mne.Epochs` parameters.
    """
    if baseline is False:
        baseline = None
    raw = ds.info['raw']
    if picks is None and raw.info['bads']:
        picks = mne.pick_types(raw.info, meg=True, eeg=True, eog=True, ref_meg=False, exclude=[])
    events = _mne_events(ds)
    out = []
    for i, tmax_i in enumerate(tmax):
        i_max = events[i, 0] + floor(tmax_i * raw.info['sfreq'])
        if raw.last_samp < i_max:
            if allow_truncation:
                tmax_i = (raw.last_samp - events[i, 0]) / raw.info['sfreq']
            else:
                missing = (i_max - raw.last_samp) / raw.info['sfreq']
                raise ValueError(f"tmax={tmax}, tmax[{i}] {tmax_i} is outside of data range by {missing:g} s")
        epochs_i = mne.Epochs(raw, events[i:i+1], None, tmin, tmax_i, baseline, picks, preload=True, **kwargs)
        out.append(epochs_i)
    return out


def raw_ndvar(raw, i_start=None, i_stop=None, decim=1, data=None, exclude='bads', sysname=None,  connectivity=None):
    """Raw data as NDVar

    Parameters
    ----------
    raw : Raw | str
        Raw instance, or path of a raw FIFF file..
    i_start : int | sequence of int
        Start sample (see notes; default is the beginning of the ``raw``).
    i_stop : int | sequence of int
        Stop sample (see notes; default is end of the ``raw``).
    decim : int
        Downsample the data by this factor when importing. ``1`` (default)
        means no downsampling. Note that this function does not low-pass filter
        the data. The data is downsampled by picking out every n-th sample.
    data : 'eeg' | 'mag' | 'grad' | None
        The kind of data to include (default based on data).
    exclude : list of string | str
        Channels to exclude (:func:`mne.pick_types` kwarg).
        If 'bads' (default), exclude channels in info['bads'].
        If empty do not exclude any.
    sysname : str
        Name of the sensor system to load sensor connectivity (e.g. 'neuromag306',
        inferred automatically for KIT data converted with a recent version of
        MNE-Python).
    connectivity : str | list of (str, str) | array of int, (n_edges, 2)
        Connectivity between elements. Can be specified as:

        - ``"none"`` for no connections
        - list of connections (e.g., ``[('OZ', 'O1'), ('OZ', 'O2'), ...]``)
        - :class:`numpy.ndarray` of int, shape (n_edges, 2), to specify
          connections in terms of indices. Each row should specify one
          connection [i, j] with i < j. If the array's dtype is uint32,
          property checks are disabled to improve efficiency.
        - ``"grid"`` to use adjacency in the sensor names

        If unspecified, it is inferred from ``sysname`` if possible.

    Returns
    -------
    data : NDVar | list of NDVar
        Data (sensor or source space). If ``i_start`` and ``i_stopr`` are scalar
        then a single NDVar is returned, if they are lists then a list of NDVars
        is returned.

    Notes
    -----
    ``i_start`` and ``i_stop`` are interpreted as event indexes (from
    :func:`mne.find_events`), i.e. relative to ``raw.first_samp``.
    """
    if not isinstance(raw, MNE_RAW):
        raw = mne_raw(raw)
    if raw.filenames and raw.filenames[0]:
        name = os.path.basename(raw.filenames[0])
    else:
        name = None
    start_scalar = i_start is None or isinstance(i_start, int)
    stop_scalar = i_stop is None or isinstance(i_stop, int)
    if start_scalar != stop_scalar:
        raise TypeError(f"i_start and i_stop must either both be scalar or both iterable, got i_start={i_start!r}, i_stop={i_stop!r}")
    elif start_scalar:
        i_start = (i_start,)
        i_stop = (i_stop,)
        scalar = True
    else:
        scalar = False

    # event index to raw index
    i_start = [i if i is None else i - raw.first_samp for i in i_start]
    i_stop = [i if i is None else i - raw.first_samp for i in i_stop]

    # target dimension
    if data is None:
        data = _guess_ndvar_data_type(raw.info)
    picks = _picks(raw.info, data, exclude)
    dim = sensor_dim(raw, picks, sysname, connectivity)
    info = _sensor_info(data, None, raw.info)

    out = []
    for start, stop in zip(i_start, i_stop):
        x = raw[picks, start:stop][0]

        if decim != 1:
            x = x[:, ::decim]
        time = UTS(0, float(decim) / raw.info['sfreq'], x.shape[1])
        out.append(NDVar(x, (dim, time), name, info))

    if scalar:
        return out[0]
    else:
        return out


def epochs_ndvar(
        epochs: Union[mne.BaseEpochs, PathArg],
        name: str = None,
        data: str = None,
        exclude: Union[str, Sequence[str]] = 'bads',
        mult: float = 1,
        info: dict = None,
        sensors: Sensor = None,
        vmax: float = None,
        sysname: str = None,
        connectivity: Union[str, Sequence] = None,
):
    """
    Convert an :class:`mne.Epochs` object to an :class:`NDVar`.

    Parameters
    ----------
    epochs
        The epochs object or path to an epochs FIFF file.
    name
        Name for the NDVar.
    data : 'eeg' | 'mag' | 'grad'
        Which data channels data to include (default based on channels in data).
    exclude
        Channels to exclude (:func:`mne.pick_types` kwarg).
        If 'bads' (default), exclude channels in info['bads'].
        If empty do not exclude any.
    mult
        multiply all data by a constant.
    info : None | dict
        Additional contents for the info dictionary of the NDVar.
    sensors
        The default (``None``) reads the sensor locations from the fiff file.
        If the fiff file contains incorrect sensor locations, a different
        Sensor can be supplied through this kwarg.
    vmax
        Set a default range for plotting.
    sysname
        Name of the sensor system to load sensor connectivity (e.g. 'neuromag306',
        inferred automatically for KIT data converted with a recent version of
        MNE-Python).
    connectivity : str | list of (str, str) | array of int, (n_edges, 2)
        Connectivity between elements. Can be specified as:

        - ``"none"`` for no connections
        - list of connections (e.g., ``[('OZ', 'O1'), ('OZ', 'O2'), ...]``)
        - :class:`numpy.ndarray` of int, shape (n_edges, 2), to specify
          connections in terms of indices. Each row should specify one
          connection [i, j] with i < j. If the array's dtype is uint32,
          property checks are disabled to improve efficiency.
        - ``"grid"`` to use adjacency in the sensor names

        If unspecified, it is inferred from ``sysname`` if possible.
    """
    if isinstance(epochs, (str, Path)):
        epochs = mne.read_epochs(epochs)

    if data is None:
        data = _guess_ndvar_data_type(epochs.info)
    picks = _picks(epochs.info, data, exclude)
    info_ = _sensor_info(data, vmax, epochs.info, info, mult)

    x = epochs.get_data()
    if len(picks) < x.shape[1]:
        x = x[:, picks]

    if mult != 1:
        x *= mult

    sensor = sensors or sensor_dim(epochs, picks, sysname, connectivity)
    time = UTS(epochs.times[0], 1. / epochs.info['sfreq'], len(epochs.times))
    return NDVar(x, ('case', sensor, time), info=info_, name=name)


def evoked_ndvar(evoked, name=None, data=None, exclude='bads', vmax=None,
                 sysname=None, connectivity=None):
    """
    Convert one or more mne :class:`Evoked` objects to an :class:`NDVar`.

    Parameters
    ----------
    evoked : str | Evoked | list of Evoked
        The Evoked to convert to NDVar. Can be a string designating a file
        path to a evoked fiff file containing only one evoked.
    name : str
        Name of the NDVar.
    data : 'eeg' | 'mag' | 'grad'
        Which data channels data to include (default based on channels in data).
    exclude : list of string | string
        Channels to exclude (:func:`mne.pick_types` kwarg).
        If 'bads' (default), exclude channels in info['bads'].
        If empty do not exclude any.
    vmax : None | scalar
        Set a default range for plotting.
    sysname : str
        Name of the sensor system to load sensor connectivity (e.g. 'neuromag306',
        inferred automatically for KIT data converted with a recent version of
        MNE-Python).
    connectivity : str | list of (str, str) | array of int, (n_edges, 2)
        Connectivity between elements. Can be specified as:

        - ``"none"`` for no connections
        - list of connections (e.g., ``[('OZ', 'O1'), ('OZ', 'O2'), ...]``)
        - :class:`numpy.ndarray` of int, shape (n_edges, 2), to specify
          connections in terms of indices. Each row should specify one
          connection [i, j] with i < j. If the array's dtype is uint32,
          property checks are disabled to improve efficiency.
        - ``"grid"`` to use adjacency in the sensor names

        If unspecified, it is inferred from ``sysname`` if possible.

    Notes
    -----
    If evoked objects have different channels, the intersection is used (i.e.,
    only the channels present in all objects are retained).
    """
    if isinstance(evoked, str):
        evoked = mne.read_evokeds(evoked)

    if isinstance(evoked, MNE_EVOKED):
        case_out = False
        evoked = (evoked,)
    elif isinstance(evoked, (tuple, list)):
        case_out = True
    else:
        raise TypeError("evoked=%s" % repr(evoked))

    # data type to load
    if data is None:
        data_set = {_guess_ndvar_data_type(e.info) for e in evoked}
        if len(data_set) > 1:
            raise ValueError("Different Evoked objects contain different "
                             "data types: %s" % ', '.join(data_set))
        data = data_set.pop()

    # MEG system
    kit_sys_ids = {e.info.get('kit_system_id') for e in evoked}
    kit_sys_ids -= {None}
    if len(kit_sys_ids) > 1:
        raise ValueError("Evoked objects from different KIT systems can not be "
                         "combined because they have different sensor layouts")
    elif kit_sys_ids:
        sysname = KIT_NEIGHBORS.get(kit_sys_ids.pop(), sysname)

    if data == 'mag':
        info = _info.for_meg(vmax)
    elif data == 'eeg':
        info = _info.for_eeg(vmax)
    elif data == 'grad':
        info = _info.for_meg(vmax, unit='T/cm')
    else:
        raise ValueError("data=%s" % repr(data))

    e0 = evoked[0]
    if len(evoked) == 1:
        picks = _picks(e0.info, data, exclude)
        x = e0.data[picks]
        if case_out:
            x = x[None, :]
        first, last, sfreq = e0.first, e0.last, e0.info['sfreq']
    else:
        # timing:  round sfreq because precision is lost by FIFF format
        timing_set = {(e.first, e.last, e.info['sfreq']) for e in evoked}
        if len(timing_set) == 1:
            first, last, sfreq = timing_set.pop()
        else:
            raise ValueError(f"Evoked objects have different timing information (first, last, sfreq): {', '.join(map(str, timing_set))}")

        # find excluded channels
        ch_sets = [set(e.info['ch_names']) for e in evoked]
        all_chs = set.union(*ch_sets)
        common = set.intersection(*ch_sets)
        exclude = set.union(*list(map(set, (e.info['bads'] for e in evoked))))
        exclude.update(all_chs.difference(common))
        exclude = list(exclude)

        # get data
        x = []
        for e in evoked:
            picks = _picks(e.info, data, exclude)
            x.append(e.data[picks])

    sensor = sensor_dim(e0, picks, sysname, connectivity)
    time = UTS.from_int(first, last, sfreq)
    if case_out:
        dims = ('case', sensor, time)
    else:
        dims = (sensor, time)
    return NDVar(x, dims, name, info)


def forward_operator(fwd, src, subjects_dir=None, parc='aparc', name=None):
    """Load forward operator as :class:`NDVar`

    Parameters
    ----------
    fwd : str | mne Forward
        MNE Forward solution, or path to forward solution.
    src : str
        Tag describing the source space (e.g., "ico-4").
    subjects_dir : str
        Location of the MRI subjects directory.
    parc : str
        Parcellation to load (corresponding to existing annot files; default
        'aparc').
    name : str
        Name the NDVar (default is the filename if a path is provided,
        otherwise "fwd").

    Returns
    -------
    fwd : NDVar  (sensor, source)
        NDVar containing the gain matrix.
    """
    is_vol = src.startswith('vol')
    if isinstance(fwd, str):
        if name is None:
            name = os.path.basename(fwd)
        fwd = mne.read_forward_solution(fwd)
        mne.convert_forward_solution(fwd, force_fixed=not is_vol, use_cps=True,
                                     copy=False)
    elif name is None:
        name = 'fwd'
    sensor = sensor_dim(fwd['info'])
    assert np.all(sensor.names == fwd['sol']['row_names'])
    subject = fwd['src'][0]['subject_his_id']
    if is_vol:
        source = VolumeSourceSpace.from_file(subjects_dir, subject, src, parc, source_spaces=fwd['src'])
        x = fwd['sol']['data'].reshape((len(sensor), len(source), 3))
        dims = (sensor, source, Space('RAS'))
    else:
        source = SourceSpace.from_file(subjects_dir, subject, src, parc, source_spaces=fwd['src'])
        x = fwd['sol']['data']
        dims = (sensor, source)
    return NDVar(x, dims, name)


def inverse_operator(inv, src, subjects_dir=None, parc='aparc', name=None):
    """Load inverse operator as :class:`NDVar`
    
    Parameters
    ----------
    inv : str | mne.minimum_norm.InverseOperator
        MNE inverse operator, or path to inverse operator.
    src : str
        Tag describing the source space (e.g., "ico-4").
    subjects_dir : str
        Location of the MRI subjects directory.
    parc : str
        Parcellation to load (corresponding to existing annot files; default
        'aparc').
    name : str
        Name the NDVar (default is the filename if a path is provided,
        otherwise "inv").

    Returns
    -------
    inv : NDVar  (source, sensor)
        NDVar containing the inverse operator.
    """
    if isinstance(inv, str):
        if name is None:
            name = os.path.basename(inv)
        inv = mne.minimum_norm.read_inverse_operator(inv)
    elif name is None:
        name = 'inv'
    sensor = sensor_dim(inv['info'], _picks(inv['info'], True, 'bads'))
    assert np.all(sensor.names == inv['eigen_fields']['col_names'])
    subject = inv['src'][0]['subject_his_id']
    source = SourceSpace.from_file(subjects_dir, subject, src, parc, source_spaces=inv['src'])
    inv = mne.minimum_norm.prepare_inverse_operator(inv, 1, 1., 'MNE')
    k = mne.minimum_norm.inverse._assemble_kernel(inv, None, 'MNE', False)[0]
    return NDVar(k, (source, sensor), {}, name)


def stc_ndvar(
        stc: Union[_BaseSourceEstimate, Sequence[_BaseSourceEstimate], PathArg],
        subject: str,
        src: str,
        subjects_dir: PathArg = None,
        method: str = None,
        fixed: bool = None,
        name: str = None,
        check: bool = True,
        parc: Union[str, None] = 'aparc',
        connectivity: str = None,
        sss_filename: str = '{subject}-{src}-src.fif',
):
    """
    Convert one or more :class:`mne.SourceEstimate` objects to an :class:`NDVar`.

    Parameters
    ----------
    stc : SourceEstimate | list of SourceEstimates | str
        The source estimate object(s) or a path to an stc file. Volum and vector
        source estimates are supported.
    subject
        MRI subject (used for loading MRI in PySurfer plotting)
    src
        The kind of source space used (e.g., 'ico-4').
    subjects_dir
        The path to the subjects_dir (needed to locate the source space file).
    method : 'MNE' | 'dSPM' | 'sLORETA' | 'eLORETA'
        Source estimation method (optional, used for generating info).
    fixed
        Source estimation orientation constraint (optional, used for generating
        info).
    name
        Ndvar name.
    check
        If multiple stcs are provided, check if all stcs have the same times
        and vertices.
    parc
        Name of a parcellation to add to the source space. ``None`` to add no
        parcellation.
    connectivity : 'link-midline'
        Modify source space connectivity to link medial sources of the two
        hemispheres across the midline.
    sss_filename
        Filename template for the MNE source space file.
    """
    subjects_dir = mne.utils.get_subjects_dir(subjects_dir)
    if isinstance(stc, Path):
        stc = str(stc)
    if isinstance(stc, str):
        stc = mne.read_source_estimate(stc)

    # construct data array
    if isinstance(stc, _BaseSourceEstimate):
        case = False
        stcs = None
    else:
        case = True
        stcs = stc
        stc = stcs[0]

    vertices = as_list(stc.vertices)  # not always list for mne < 0.21
    if stcs and check:
        times = stc.times
        for stc_ in stcs[1:]:
            assert np.array_equal(stc_.times, times)
            stc_vertices = as_list(stc_.vertices)
            for v1, v0 in zip_longest(stc_vertices, vertices):
                assert np.array_equal(v1, v0)

    if stcs:
        x = np.array([s.data for s in stcs])
    else:
        x = stc.data

    # Construct NDVar Dimensions
    time = UTS(stc.tmin, stc.tstep, stc.times.size)
    if isinstance(stc, MNE_VOLUME_STC):
        ss = VolumeSourceSpace(vertices, subject, src, subjects_dir, None, filename=sss_filename)
        is_vector = stc.data.ndim == 3
    elif isinstance(stc, (mne.SourceEstimate, mne.VectorSourceEstimate)):
        ss = SourceSpace(vertices, subject, src, subjects_dir, parc, filename=sss_filename)
        is_vector = isinstance(stc, mne.VectorSourceEstimate)
    else:
        raise TypeError(f"stc={stc!r}")
    # Apply connectivity modification
    if isinstance(connectivity, str):
        if connectivity == 'link-midline':
            ss._link_midline()
        elif connectivity != '':
            raise ValueError(f"connectivity={connectivity!r}")
    elif connectivity is not None:
        raise TypeError(f"connectivity={connectivity!r}")
    # assemble dims
    dims = [ss, time]
    if is_vector:
        dims.insert(1, Space('RAS'))
    if case:
        dims.insert(0, Case)

    # find the right measurement info
    info = {}
    if fixed is False:
        info['meas'] = 'Activation'
        if method:
            info['unit'] = method
    elif fixed is True:
        info['meas'] = 'Current Estimate'
        if method == 'MNE':
            info['unit'] = 'A'
        elif method:
            info['unit'] = method
    elif fixed is not None:
        raise ValueError(f"fixed={fixed!r}")

    return NDVar(x, dims, name, info)


def _trim_ds(ds, epochs):
    """Trim a Dataset to account for rejected epochs.

    If no epochs were rejected, the original ds is rturned.

    Parameters
    ----------
    ds : Dataset
        Dataset that was used to construct epochs.
    epochs : Epochs
        Epochs loaded with mne_epochs()
    """
    if len(epochs) < ds.n_cases:
        ds = ds.sub(epochs.selection)
        ds.info['epochs.selection'] = epochs.selection

    return ds
